from collections.abc import Iterable
from pathlib import Path
from typing import Any, Union

import geopandas as gpd
import numpy as np
import xarray as xr
from cloudpathlib import CloudPath
from rasterio.io import DatasetReader, DatasetWriter
from shapely import MultiPolygon, Polygon

AnyPathType = Union[CloudPath, Path]
"""Any Path Type (derived from Pathlib and CloudpathLib)"""

AnyPathStrType = Union[str, CloudPath, Path]
"""Same as :code:`AnyPathType` but appened with :code:`str`"""

AnyXrDataStructure = Union[xr.DataArray, xr.Dataset]
"""Xarray's DataArray or Dataset"""

AnyNumpyArray = Union[np.ndarray, np.ma.masked_array]
"""Numpy array or masked array"""

AnyPolygonType = Union[Polygon, MultiPolygon]
"""Shapely Polygon or MultiPolygon"""

AnyRioDatasetType = Union[DatasetReader, DatasetWriter]
""" Any Rasterio Dataset Type (both Reader, Writer) """

AnyRasterType = Union[
    AnyPathStrType, tuple[AnyNumpyArray, dict], AnyXrDataStructure, AnyRioDatasetType
]
"""
Any object potentially describing a raster:

- its path,
- its ``xarray`` representation (:class:`xarray:xarray.Dataset` or :class:`xarray:xarray.DataArray`),
- its ``rasterio`` representation (``DatasetReader`` or ``DatasetWriter``)
- or its array + metadata (``np.ndarray`` + dict)
"""

AnyVectorType = Union[AnyPathStrType, gpd.GeoDataFrame]
"""
Any object potentially describing a vector:

- its path,
- its :class:`geopandas:geopandas.GeoDataFrame`
"""


def is_iterable(obj: Any, str_allowed: bool = False):
    """
    Is the object an iterable?

    Useful to replace this kind of code:

        >>> if isinstance(my_items, (list, tuple)):
        >>>     do...

    by:

        >>> if is_iterable(my_items):
        >>>     do...

    It allows not to forget checking for some iterables you aren't aware of.

    Args:
        obj (Any): Object to check
        str_allowed (bool): If set, strings are considered as iterable. If not, they are not considered as iterable.

    Returns:
        bool: True if the iobject is iterable

    Examples:

        >>> is_iterable((1, 2, 3))
        True
        >>> is_iterable([1, 2, 3])
        True
        >>> is_iterable({1, 2, 3})
        True
        >>> is_iterable(np.array([1, 2, 3]))
        True
        >>> is_iterable("1, 2, 3")
        False
        >>> is_iterable("1, 2, 3", str_allowed=True)
        True
        >>> is_iterable(1)
        False
        >>> is_iterable(AnyPath("1, 2, 3"))
        False
    """
    if isinstance(obj, str) and not str_allowed:
        return False
    else:
        return isinstance(obj, Iterable)


def make_iterable(
    obj: Any, str_allowed: bool = False, convert_none: bool = False
) -> list:
    """
    Convert the object to a list if this object is not iterable

    Useful to replace this kind of code:

        >>> if to_convert is not None and not isinstance(to_convert, (list, tuple)):
        >>>    to_convert = [to_convert]

    by:

        >>> to_convert = make_iterable(to_convert)

    or:

        >>> if isinstance(my_items, (list, tuple)):
        >>>    first_item = my_items[0]
        >>> else:
        >>>    first_item = my_items

    by:

        >>> first_item = make_iterable(to_convert)[0]

    Args:
        obj (Any): Object to check
        str_allowed (bool): If set, strings are considered as iterable. If not, they are not considered as iterable.
        convert_none (bool): If true, if obj is None, then it won't be converted into a list. By default, Nones are not converted to list.

    Returns:
        list: Object as an iterable

    Examples:

        >>> make_interable((1, 2, 3))
        (1, 2, 3)
        >>> make_interable([1, 2, 3])
        [1, 2, 3]
        >>> make_interable({1, 2, 3})
        {1, 2, 3}
        >>> make_interable(np.array([1, 2, 3]))
        np.array([1, 2, 3])
        >>> make_interable("1, 2, 3", str_allowed=True)
        "1, 2, 3"
        >>> make_interable("1, 2, 3", str_allowed=False)
        ["1, 2, 3"]
        >>> make_interable(1)
        [1]
        >>> make_interable(AnyPath("1, 2, 3"))
        [AnyPath("1, 2, 3")]
        >>> make_interable(None)
        None
        >>> make_interable(None, convert_none=True)
        [None]

    """
    if (convert_none and obj is None) or not is_iterable(obj, str_allowed):
        obj = [obj]

    return obj



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing the miscellaneous functions"""

import os

import pytest

from ci.script_utils import Polarization
from sertit import __version__, ci, misc
from sertit.misc import compare, compare_version

ci.reduce_verbosity()


def test_run_command():
    """Test run_command"""
    cmd = "cd .."
    misc.run_cli(
        cmd, in_background=False, cwd="/"
    )  # Just ensure no exception is thrown
    cmd = ["cd", ".."]
    misc.run_cli(cmd, in_background=True, cwd="/")  # Just ensure no exception is thrown


def test_get_function_name():
    """Test get_function_name"""
    ci.assert_val(
        misc.get_function_name(), "test_get_function_name", "get_function_name"
    )


def test_in_docker():
    """Test in_docker"""
    # Hack: in docker if ni linux
    misc.in_docker()  # Just hope it doesn't crash


def test_chdir():
    """Testing chdir functions"""
    curr_dir = os.path.dirname(os.path.realpath(__file__))
    old_pwd = os.getcwd()
    with misc.chdir(curr_dir):
        pwd = os.getcwd()
        assert pwd == curr_dir

    ci.assert_val(os.getcwd(), old_pwd, "test_chdir")


def test_select_prune_dict():
    keys = ["a", "b", "c"]
    d = {"a": 1, "b": 2, "c": 3, "d": 4}

    # Select dict
    ci.assert_val(misc.select_dict(d, keys), {"a": 1, "b": 2, "c": 3}, "Select dict")

    # Prune dict
    ci.assert_val(misc.prune_dict(d, keys), {"d": 4}, "Prune dict")


def test_list_dict():
    """Test dict functions"""
    test_list = ["A", "T", "R", "", 3, None, "0"]
    test_dict = {"A": "T", "R": 3, "B": None, "C": "", "D": "0"}

    # Remove empty values (list)
    test_list = misc.remove_empty_values(test_list, other_empty_values=["0"])
    ci.assert_val(test_list, ["A", "T", "R", 3], "Remove empty values from list")

    # Remove empty values (dict)
    test_dict = misc.remove_empty_values(test_dict, other_empty_values=["", "0"])
    ci.assert_val(test_dict, {"A": "T", "R": 3}, "Remove empty values from dict")

    # List to dict
    ci.assert_val(misc.list_to_dict(test_list), test_dict, "List to dict")

    # Nested set
    res_dict = {"A": "T", "R": 3, "B": {"C": {"D": "value"}}}
    misc.nested_set(test_dict, ["B", "C", "D"], "value")
    ci.assert_val(test_dict, res_dict, "Nested set")

    # Mandatory keys
    misc.check_mandatory_keys(test_dict, ["A", "B"])  # True
    with pytest.raises(ValueError):
        misc.check_mandatory_keys(test_dict, ["C"])  # False

    # Find by key
    ci.assert_val(misc.find_by_key(test_dict, "D"), "value", "Find by key")


def test_enum():
    """Test ListEnum"""

    ci.assert_val(Polarization.list_values(), ["HH", "VV", "VH", "HV"], "Values")
    ci.assert_val(Polarization.list_names(), ["hh", "vv", "vh", "hv"], "Names")
    ci.assert_val(Polarization.from_value("HH"), Polarization.hh, "From string value")
    ci.assert_val(
        Polarization.from_value(Polarization.hh), Polarization.hh, "From enum value"
    )

    with pytest.raises(ValueError):
        Polarization.from_value("ZZ")

    ci.assert_val(
        Polarization.convert_from(["HH", "vv", Polarization.hv]),
        [
            Polarization.hh,
            Polarization.vv,
            Polarization.hv,
        ],
        "convert_from",
    )


def test_unique():
    """Test unique function"""
    non_unique = [1, 2, 20, 6, 210, 2, 1]
    unique = [1, 2, 20, 6, 210]
    ci.assert_val(unique, misc.unique(non_unique), "Unique")


def test_comparisons():
    """Test comparisons"""
    # -- True --
    # Equalities
    assert compare(1, 1, "==")
    assert compare(1, 1, ">=")
    assert compare(1, 1, "<=")

    # Inequalities
    assert compare(1, 2, "<")
    assert compare(1, 0.5, ">")

    # -- False --
    # Equalities
    assert not compare(1, 2, "==")
    assert not compare(1, 5, ">=")
    assert not compare(1, 0.5, "<=")

    # Inequalities
    assert not compare(1, 0.5, "<")
    assert not compare(1, 2, ">")


def test_compare_versions():
    """"""
    assert compare_version("geopandas", "0.10.0", ">=")
    assert compare_version("geopandas", "5.0.0", "<")
    assert compare_version(__version__, "5.0.0", "<")
    assert compare_version(__version__, "1.0.0", ">")



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Geometry tools

You can use this only if you have installed sertit[full] or sertit[vectors]
"""

import logging

import geopandas as gpd
import numpy as np
import shapely
from shapely import ops
from shapely.errors import GeometryTypeError
from shapely.geometry import Polygon, box
from tqdm import tqdm

from sertit import misc, vectors
from sertit.logs import SU_NAME
from sertit.types import AnyPolygonType

LOGGER = logging.getLogger(SU_NAME)


def from_polygon_to_bounds(polygon: AnyPolygonType) -> (float, float, float, float):
    """
    Convert a :code:`shapely.polygon` to its bounds, sorted as :code:`left, bottom, right, top`.

    Args:
        polygon (MultiPolygon): polygon to convert

    Returns:
        (float, float, float, float): left, bottom, right, top

    Example:
        >>> poly = Polygon(((0., 0.), (0., 1.), (1., 1.), (1., 0.), (0., 0.)))
        >>> from_polygon_to_bounds(poly)
        (0.0, 0.0, 1.0, 1.0)
    """
    left = polygon.bounds[0]  # xmin
    bottom = polygon.bounds[1]  # ymin
    right = polygon.bounds[2]  # xmax
    top = polygon.bounds[3]  # ymax

    assert left < right
    assert bottom < top

    return left, bottom, right, top


def from_bounds_to_polygon(
    left: float, bottom: float, right: float, top: float
) -> Polygon:
    """
    Convert the bounds to a :code:`shapely.polygon`.

    Args:
        left (float): Left coordinates
        bottom (float): Bottom coordinates
        right (float): Right coordinates
        top (float): Top coordinates

    Returns:
        Polygon: Polygon corresponding to the bounds

    Example:
        >>> poly = from_bounds_to_polygon(0.0, 0.0, 1.0, 1.0)
        >>> print(poly)
        'POLYGON ((1 0, 1 1, 0 1, 0 0, 1 0))'

    """
    return box(min(left, right), min(top, bottom), max(left, right), max(top, bottom))


def get_wider_exterior(vector: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """
    Get the wider exterior of a MultiPolygon as a Polygon

    Args:
        vector (gpd.GeoDataFrame): Polygon to simplify

    Returns:
        vector: gpd.GeoDataFrame: Wider exterior

    Example:
        >>> # Open a raw footprint
        >>> footprint_raw = vectors.read("footprint_raw.geojson")
                                                           geometry
        0         MULTIPOLYGON (((491053.524 5616778.498, 491262...
        1         MULTIPOLYGON (((491314.496 5616444.620, 491295...
        2         MULTIPOLYGON (((490783.440 5616102.457, 490923...
        >>>
        >>> # Get the wider exterior
        >>> get_wider_exterior(footprint_raw)
                                                        geometry
        0      POLYGON ((491053.524 5616778.498, 491262.302 5...
    """
    vector = vector.explode(index_parts=True)

    # Get the footprint max (discard small holes stored in other polygons)
    wider = vector[vector.area == np.max(vector.area)]

    # Only select the exterior of this footprint(sometimes some holes persist)
    if not wider.empty:
        poly = Polygon(list(wider.exterior.iat[0].coords))
        wider = gpd.GeoDataFrame(geometry=[poly], crs=wider.crs)

        # Resets index as we only got one polygon left which should have index 0
        wider.reset_index(inplace=True)

    return wider


def make_valid(gdf: gpd.GeoDataFrame, verbose=False) -> gpd.GeoDataFrame:
    """
    Repair geometries from a dataframe.

    Better to use :code:`gpd.make_valid` if you can.

    Args:
        gdf (gpd.GeoDataFrame): GeoDataFrame to repair
        verbose (bool): Verbose invalid geometries

    Returns:
        gpd.GeoDataFrame: Repaired geometries

    Example:
        >>> # Open a raw  vector with invalid geometries
        >>> raw = vectors.read("raw.geojson")
                                                           geometry
        0         MULTIPOLYGON (((491053.524 5616778.498, 491262...
        1         MULTIPOLYGON (((491314.496 5616444.620, 491295...
        2         MULTIPOLYGON (((490783.440 5616102.457, 490923...
        >>>
        >>> # Get the valid geometries
        >>> make_valid(raw)
                                                           geometry
        1         MULTIPOLYGON (((491314.496 5616444.620, 491295...
    """
    try:
        geos_logger = logging.getLogger("shapely.geos")
        previous_level = geos_logger.level
        if verbose:
            logging.debug(f"Invalid geometries:\n\t{gdf[~gdf.is_valid]}")
        else:
            geos_logger.setLevel(logging.CRITICAL)

        # Discard self-intersection and null geometries
        from shapely.validation import make_valid

        gdf.geometry = gdf.geometry.apply(make_valid)

        if not verbose:
            geos_logger.setLevel(previous_level)
    except ImportError:
        import shapely

        LOGGER.warning(
            f"'make_valid' not available in 'shapely' (version {shapely.__version__} < 1.8). "
            f"The obtained vector may be broken !"
        )

    return gdf


def simplify_footprint(
    footprint: gpd.GeoDataFrame, resolution: float, max_nof_vertices: int = 50
) -> gpd.GeoDataFrame:
    """
    Simplify footprint.

    Set a number of maximum vertices and this function will try to simplify the footprint to have less than this number of vertices.
    The tolerance will grow to try to respect this number of vertices.

    This function will loop over a number of pixels of tolerence [1, 2, 4, 8, 16, 32, 64] (tolerance of gpd.simplify == resolution * tol_pix)
    If in the end, the number of vertices is still too high, a warning will be emitted.

    Args:
        footprint (gpd.GeoDataFrame): Footprint to be simplified
        resolution (float): Corresponding resolution
        max_nof_vertices (int): Maximum number of vertices of the wanted footprint

    Returns:
        gpd.GeoDataFrame: Simplified footprint

    Examples:
        >>> # Open a raw footprint
        >>> footprint_raw = vectors.read("footprint_raw.geojson")
        >>> len(footprint_raw.get_coordinates())
        64757
        >>>
        >>> # Get the simplified footprint
        >>> simplified = simplify_footprint(footprint_raw, 20)
        >>> len(simplified.get_coordinates())
        29
    """
    # Number of pixels of tolerance
    tolerance = [1, 2, 4, 8, 16, 32, 64, 128, 256]

    # Process only if given footprint is too complex (too many vertices)
    def simplify_geom(value):
        nof_vertices = len(value.exterior.coords)
        if nof_vertices > max_nof_vertices:
            for tol in tolerance:
                # Simplify footprint
                value = value.simplify(
                    tolerance=tol * resolution, preserve_topology=True
                )

                # Check if OK
                nof_vertices = len(value.exterior.coords)
                if nof_vertices <= max_nof_vertices:
                    break

        # WARNING if nof_vertices > max_nof_vertices
        nof_vertices = len(value.exterior.coords)
        if nof_vertices > max_nof_vertices:
            LOGGER.warning(
                f"The number of vertices ({nof_vertices}) of your simplified footprint is higher than {max_nof_vertices}."
                f"However, it cannot be simplified further according to the given resolution ({resolution})."
            )

        return value

    footprint = footprint.explode(index_parts=True)
    footprint.geometry = footprint.geometry.apply(simplify_geom)

    return footprint


def fill_polygon_holes(
    gpd_results: gpd.GeoDataFrame, threshold: float = None
) -> gpd.GeoDataFrame:
    """
    Fill holes over a given threshold on the hole area (in meters) for all polygons of a GeoDataFrame.
    If the threshold is set to None, every hole is filled.

    Args:
        gpd_results (gpd.GeoDataFrame): Geodataframe filled whith drilled polygons
        threshold (float): Holes area threshold, in meters. If set to None, every hole is filled.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with filled holes

    Example:
        >>> # Open a polygon with holes
        >>> holes = vectors.read("holes.geojson")
        >>> # This polygons has interiors features, it has holes
        >>> holes.interiors
        3    [LINEARRING (491328.9981955575 5616655.8234532...
        dtype: object
        >>> no_holes = fill_polygon_holes(holes)
        Processing objects: 100%|██████████| 1/1 [00:00<00:00, 897.18it/s]
        >>> no_holes.interiors
        0    []
        dtype: object
    """

    def _fill_polygon(polygon: Polygon, threshold: float = None):
        """
        Function used in apply in fill_polygon_holes
        """
        if threshold is not None:
            if threshold > 0 and len(polygon.interiors) > 0:
                new_interiors = list(polygon.interiors)
                for interior in polygon.interiors:
                    interior_poly = Polygon(np.array(interior.coords))
                    if interior_poly.area < threshold:
                        new_interiors.remove(interior)
                return Polygon(polygon.exterior, new_interiors)
            else:
                return polygon
        else:
            return Polygon(polygon.exterior)

    # Ensure the geometries are valid
    gpd_results = make_valid(gpd_results)

    # Check if vector is projected, if not convert it
    with vectors.utm_crs(gpd_results) as utm_results:
        # Explode the multipolygons
        utm_results = utm_results.explode(index_parts=False)

        # Keep only the exterior
        tqdm.pandas(desc="Processing objects")
        utm_results.geometry = utm_results.geometry.progress_apply(
            _fill_polygon, args=(threshold,)
        )

        # Dissolve and explode to remove polygons in polygons
        utm_results = utm_results.dissolve()
        gpd_results = utm_results.explode(index_parts=False)

    # Write back to file
    return gpd_results


def line_merge(lines: gpd.GeoDataFrame, **kwargs) -> gpd.GeoDataFrame:
    """
    :code:`shapely.line_merge` algorithm applied to a GeoDataFrame.

    See the corresponding documentation for more insights about the details of this function.

    Args:
        lines (gpd.GeoDataFrame): MultiLineString as a GeoDataFrame.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame composed of (Multi)LineStrings formed by combining the lines of the input GeoDataFrame

    Example:
        >>> import geopandas as gpd
        >>> from sertit import geometry
        >>> lines = gpd.read("my_lines.shp")
        >>> poly = gpd.read("my_poly.shp")
                                                    geometry
        0  POLYGON ((491460.248 5616687.073, 491460.248 5...
        >>> geometry.split(poly, splitter=lines)
        The lines are discontinuous and don't intersect totally the input polygon: nothing is splitted
                                                    geometry
        0  POLYGON ((491460.248 5616687.073, 491460.248 5...

        >>> lines = geometry.line_merge(lines)
        >>> geometry.split(poly, splitter=lines)
        The lines are now merged into one and now intersect totally the input polygon: split is done
                                                            geometry
        0  POLYGON ((491460.248 5616687.073, 491460.248 5...
        0  POLYGON ((491055.017 5616255.823, 491053.998 5...
    """
    merge_lines = shapely.line_merge(lines.dissolve().geometry.values, **kwargs)
    return gpd.GeoDataFrame(geometry=merge_lines, crs=lines.crs).explode(
        ignore_index=True
    )


def split(polygons: gpd.GeoDataFrame, splitter: gpd.GeoDataFrame):
    """
    Split polygons with polygons or lines.

    :code:`shapely.ops.split` algorithm applied to GeoDataFrames.

    Be careful: lines have to cut the whole polygon to work!
    Use :code:`geometry.line_merge: to merge your lines if needed.

    Args:
        polygons (gpd.GeoDataFrame): Polygons to split
        splitter (gpd.GeoDataFrame): Splitter to split the polygons

    Returns:
        gpd.GeoDataFrame: Split GeoDataFrame

    Example:
        >>> import geopandas as gpd
        >>> from sertit import geometry
        >>> lines = gpd.read("my_lines.shp")
        >>> poly = gpd.read("my_poly.shp")
                                                    geometry
        0  POLYGON ((491460.248 5616687.073, 491460.248 5...
        >>> split_poly = geometry.split(poly, splitter=lines)
                                                            geometry
        0  POLYGON ((491460.248 5616687.073, 491460.248 5...
        0  POLYGON ((491055.017 5616255.823, 491053.998 5...
    """
    out = polygons.dropna(axis=1).geometry
    for _, split in splitter.iterrows():
        # Compute the boundary of the splitter polygon (to get a LineString)
        if split.geometry.area > 0:
            boundary = split.geometry.boundary
        else:
            boundary = split.geometry

        # Explode to prevent FeatureCollections
        try:
            # LineStrings
            out = (
                out.map(lambda geom: ops.split(geom, boundary).geoms).explode().dropna()  # noqa: B023
            )
        except GeometryTypeError:
            # MultiLineStrings
            for line in boundary.geoms:
                out = (
                    out.map(lambda geom: ops.split(geom, line).geoms).explode().dropna()  # noqa: B023
                )

    return gpd.GeoDataFrame(geometry=out.explode(), crs=polygons.crs)


def intersects(
    input: gpd.GeoDataFrame, other: gpd.GeoDataFrame, buffer_on_input: float = None
) -> gpd.GeoDataFrame:
    """
    Select the polygons of the input GeoDataFrame that intersects the other one and return them.

    A buffer can be added on the input GeoDataFrame to be sure the intersection is ok.
    For example, when dealing with polygons and points on their borders, sometimes the points fall a tiny bit outside the polygon boundary and is not considred as intersecting.

    :code:`gpd.intersects` algorithm applied to whole GeoDataFrames.

    Args:
        input (gpd.GeoDataFrame): Input GeoDataFrame from which the polygons will be selected
        other (gpd.GeoDataFrame): Other GeoDataFrame from that will intersect the first one
        buffer_on_input (float): Buffer size on thhe input GeoDataFrame

    Returns:
        gpd.GeoDataFrame: Polygons of the input that intersects the other GeoDataFrame

    Examples:
        >>> water = vectors.read("water.geojson")
        >>> lakes = vectors.read("lakes.geojson")
        >>> intersects(water, lakes)
                                                    geometry
        2  POLYGON ((490733.035 5616749.035, 490936.972 5...
        3  POLYGON ((491254.800 5616242.894, 491175.035 5...

        >>> water = vectors.read("water.geojson")
        >>> samples = vectors.read("samples.geojson")
        >>>
        >>> # Some samples are on the polygon boundary but for some reason fall a tiny bit outside and don't intersect the water
        >>> intersects(water, samples)
                                                    geometry
        2  POLYGON ((490733.035 5616749.035, 490936.972 5...
        >>>
        >>> # Some samples are on the polygon boundary but for some reason fall a tiny bit outside and don't intersect the water
        >>> # Use the buffer to handle this case. Use it carefully!
        >>> intersects(water, samples, buffer_on_input=0.1)
                                                    geometry
        2  POLYGON ((490733.035 5616749.035, 490936.972 5...
        3  POLYGON ((491254.800 5616242.894, 491175.035 5...
    """
    if buffer_on_input is not None:
        input_to_intersect = buffer(input, buffer_on_input)
    else:
        input_to_intersect = input

    return input[
        input_to_intersect.geometry.map(lambda x: x.intersects(other.geometry).any())
    ]


def buffer(vector: gpd.GeoDataFrame, buffer_m: float, **kwargs) -> gpd.GeoDataFrame:
    """
    Add a buffer on a vector.

    :code:`gpd.buffer` algorithm returning a GeoDataFrame instead of a GeoSeries.

    Args:
        vector (gpd.GeoDataFrame): Input vector
        buffer_m (int): Buffer size in meters.
        **kwargs: Other buffer arguments

    Returns:
        gpd.GeoDataFrame: Buffered vector

    Example:
        >>> lines = vectors.read(r"lines.shp")
        >>> lines.area
        0    0.0
        1    0.0
        dtype: float64
        >>> lines_buffer = buffer(lines, 10)
        >>> lines_buffer.area
        0    5695.213033
        1    5898.723719
        dtype: float64

    """
    vector_bfd = vector.copy()
    vector_bfd.geometry = vector.buffer(buffer_m, **kwargs)
    return vector_bfd


def nearest_neighbors(
    src_gdf: gpd.GeoDataFrame,
    candidates_gdf: gpd.GeoDataFrame,
    method: str,
    k_neighbors: int = None,
    radius: float = None,
    **kwargs,
) -> (np.ndarray, np.ndarray):
    """
    For each point in src_gdf, find the closest point in candidates_gdf and return them with their distances (in the crs coordinates).

    Closest points are:

    - if method == :code:`k_neighbors`: the k closest neighbors
    - if method == :code:`radius`: the neighbors inside this radius (in the crs coordinates, better done with projected geometries)

    Args:
        src_gdf (gpd.GeoDataFrame): Source geodataframe
        candidates_gdf (gpd.GeoDataFrame): Candidates geodataframe
        method (str): 'k_neighbors' or 'radius'
        k_neighbors (int): Number of neighbors to be looked for
        radius (float): Radius in which to find the neighbors
        **kwargs: Other args for the query

    Returns:
        (np.ndarray, np.ndarray): closest samples, distances

    Examples:
        >>> from sertit import geometry, vectors
        >>> src = vectors.read("src.shp")
        >>> candidates = vectors.read("candidates.shp")
        >>> # There is only one point in the neighborhood of each src, the others are further than 100m
        >>>
        >>> # Radius method
        >>> nearest_neighbors(src, candidates, method="radius", radius=100)
        [array([13]) array([12]) array([0])], [array([39.62574458]) array([50.37121574]) array([90.98648454])]
        >>>
        >>> # k_neighbors method
        >>> nearest_neighbors(src, candidates, method="k_neighbors", k_neighbors=1)
        [array([13]) array([12]) array([0])], [array([39.62574458]) array([50.37121574]) array([90.98648454])]
    """
    # Parse coordinates from points and insert them into a numpy array as RADIANS
    src = [(val.xy[0][0], val.xy[1][0]) for val in src_gdf.geometry.values]
    candidates = [
        (val.xy[0][0], val.xy[1][0]) for val in candidates_gdf.geometry.values
    ]

    # Find the nearest points
    # -----------------------
    # closest ==> index in right_gdf that corresponds to the closest point
    # dist ==> distance between the nearest neighbors (in the crs coordinates)
    if method == "k_neighbors":
        closest_samples, distances = _get_k_nearest(
            src_points=src, candidates=candidates, k_neighbors=k_neighbors, **kwargs
        )
    else:
        closest_samples, distances = _get_radius_nearest(
            src_points=src, candidates=candidates, radius=radius, **kwargs
        )

    return closest_samples, distances


def _get_k_nearest(src_points: list, candidates: list, k_neighbors: int, **kwargs):
    """
    For each point in src_gdf, find the nearest k points in candidates_gdf and return them with their distance.

    Args:
        src_points (list): Source points
        candidates (list): Candidate points
        k_neighbors (int): Number of neighbors to be looked for
        **kwargs: Other args for the query

    Returns:
        (np.ndarray, np.ndarray): closest samples, distances
    """
    try:
        from sklearn.neighbors import BallTree
    except ModuleNotFoundError as ex:
        raise ModuleNotFoundError(
            "Please install 'sklearn' the 'geometry.nearest_neighbors' function."
        ) from ex

    # Create tree from the candidate points
    tree = BallTree(candidates, leaf_size=15)

    # Find the closest points and distances
    closest_dist, closest = tree.query(
        src_points, k=min(k_neighbors, len(candidates)), **kwargs
    )

    # Return indices and distances
    return closest, closest_dist


def _get_radius_nearest(
    src_points: list, candidates: list, radius: float, **kwargs
) -> (np.ndarray, np.ndarray):
    """
    For each point in src_gdf, find the points in candidates_gdf inside the given radius and return them with their distance.

    Args:
        src_points (list): Source points
        candidates (list): Candidate points
        radius (float): Radius for the search
        **kwargs: Other args for the query

    Returns:
        (np.ndarray, np.ndarray): closest samples, distances
    """
    try:
        from sklearn.neighbors import BallTree
    except ModuleNotFoundError as ex:
        raise ModuleNotFoundError(
            "Please install 'sklearn' the 'geometry.nearest_neighbors' function."
        ) from ex

    # Create tree from the candidate points
    tree = BallTree(candidates, leaf_size=15)

    # Find the closest points and distances
    closest, closest_dist = tree.query_radius(
        src_points,
        r=radius,
        return_distance=True,
        **misc.select_dict(kwargs, ["count_only", "sort_results"]),
    )

    # Return indices and distances
    return closest, closest_dist


def force_2d(vect: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Have the force_2d function even with geopandas < 1.0.0"""
    try:
        return vect.force_2d()
    except AttributeError:
        vect.geometry = vect.geometry.apply(shapely.force_2d)
        return vect


def force_3d(vect: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Have the force_3d function even with geopandas < 1.0.0"""
    try:
        return vect.force_3d()
    except AttributeError:
        vect.geometry = vect.geometry.apply(shapely.force_3d)
        return vect



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing raster function (with rasterio)"""

import os
import shutil

import numpy as np
import pytest
import rasterio
import shapely
from rasterio.windows import Window

from ci.script_utils import KAPUT_KWARGS, rasters_path, s3_env
from sertit import ci, rasters_rio, vectors
from sertit.rasters_rio import any_raster_to_rio_ds, get_nodata_value_from_dtype
from sertit.vectors import EPSG_4326

ci.reduce_verbosity()


@pytest.fixture
def raster_path():
    return rasters_path().joinpath("raster.tif")


@pytest.fixture
def raster_meta(raster_path):
    return rasters_rio.read(raster_path, **KAPUT_KWARGS)


@pytest.fixture
def mask_path():
    return rasters_path().joinpath("raster_mask.geojson")


@pytest.fixture
def mask(mask_path):
    return vectors.read(mask_path)


@pytest.fixture
def ds_dtype(raster_path):
    with rasterio.open(str(raster_path)) as ds:
        return ds.meta["dtype"]


@s3_env
def test_read(tmp_path, raster_path, mask_path, raster_meta):
    """Test read functions"""
    raster_window_path = rasters_path().joinpath("window.tif")
    raster_window_20_path = rasters_path().joinpath("window_20.tif")
    extent_path = rasters_path().joinpath("extent.geojson")
    footprint_path = rasters_path().joinpath("footprint.geojson")

    # Get Extent
    extent = rasters_rio.get_extent(raster_path)
    truth_extent = vectors.read(extent_path)
    ci.assert_geom_equal(extent, truth_extent)

    # Get Footprint
    footprint = rasters_rio.get_footprint(raster_path)
    truth_footprint = vectors.read(footprint_path)
    ci.assert_geom_equal(footprint, truth_footprint)

    # Read
    raster, meta = raster_meta
    with rasterio.open(str(raster_path)) as ds:
        raster_1, meta1 = rasters_rio.read(ds, resolution=20)
        raster_2, _ = rasters_rio.read(ds, resolution=[20, 20])
        raster_3, _ = rasters_rio.read(ds, size=(meta1["width"], meta1["height"]))
        raster_4, _ = rasters_rio.read(raster_path, indexes=[1])
        with pytest.raises(ValueError):
            rasters_rio.read(ds, resolution=[20, 20, 20])

        assert raster.shape == (ds.count, ds.height, ds.width)
        assert meta["crs"] == ds.crs
        assert meta["transform"] == ds.transform
        np.testing.assert_array_equal(raster_1, raster_2)
        np.testing.assert_array_equal(raster_1, raster_3)
        np.testing.assert_array_equal(raster, raster_4)  # 2D array

        # -- Read with window
        window_out = os.path.join(tmp_path, "test_xda_window.tif")
        window, w_mt = rasters_rio.read(
            raster_path,
            window=mask_path,
        )
        rasters_rio.write(window, w_mt, window_out, **KAPUT_KWARGS)
        ci.assert_raster_equal(window_out, raster_window_path)

        # Gdf
        window_20_out = os.path.join(tmp_path, "test_xda_20_window.tif")
        gdf = vectors.read(mask_path)
        bounds = gdf.bounds.values[0]
        window_20, w_mt_20 = rasters_rio.read(
            raster_path, window=gdf.to_crs(EPSG_4326), resolution=20
        )
        rasters_rio.write(window_20, w_mt_20, window_20_out)
        ci.assert_raster_equal(window_20_out, raster_window_20_path)

        # Bounds
        window_20_2, w_mt_20_2 = rasters_rio.read(
            raster_path, window=bounds, resolution=20
        )
        np.testing.assert_array_equal(window_20, window_20_2)
        ci.assert_meta(w_mt_20, w_mt_20_2)

        # Window
        window_20_3, w_mt_20_3 = rasters_rio.read(
            raster_path,
            window=Window(col_off=57, row_off=0, width=363, height=321),
            resolution=20,
        )
        np.testing.assert_array_equal(window_20, window_20_3)
        ci.assert_meta(w_mt_20, w_mt_20_3)

        with pytest.raises(FileNotFoundError):
            rasters_rio.read(
                raster_path,
                window=rasters_path().joinpath("non_existing_window.kml"),
            )


@s3_env
def test_write(tmp_path, raster_path, raster_meta):
    """Test write functions"""
    raster_out = os.path.join(tmp_path, "test.tif")

    rasters_rio.write(*raster_meta, raster_out)

    assert os.path.isfile(raster_out)
    ci.assert_raster_equal(raster_path, raster_out)


@s3_env
def test_mask(tmp_path, raster_path, mask):
    """Test mask function"""
    raster_masked_path = rasters_path().joinpath("raster_masked.tif")
    raster_masked_out = os.path.join(tmp_path, "test_mask.tif")

    with rasterio.open(str(raster_path)) as ds:
        mask_arr, mask_meta = rasters_rio.mask(ds, mask.geometry)

    rasters_rio.write(mask_arr, mask_meta, raster_masked_out)
    ci.assert_raster_equal(raster_masked_out, raster_masked_path)


@s3_env
def test_crop(tmp_path, raster_path, mask):
    """Test crop function"""
    # Crop
    raster_cropped_path = rasters_path().joinpath("raster_cropped.tif")
    raster_cropped_out = os.path.join(tmp_path, "test_crop.tif")

    with rasterio.open(str(raster_path)) as ds:
        crop_arr, crop_meta = rasters_rio.crop(ds, mask)

    rasters_rio.write(crop_arr, crop_meta, raster_cropped_out)
    ci.assert_raster_equal(raster_cropped_out, raster_cropped_path)


@s3_env
def test_sieve(tmp_path, raster_meta):
    """Test sieve function"""
    raster_sieved_path = rasters_path().joinpath("raster_sieved.tif")

    # Sieve
    raster, meta = raster_meta
    sieve_out = os.path.join(tmp_path, "test_sieved.tif")

    sieve_arr, sieve_meta = rasters_rio.sieve(
        raster, meta, sieve_thresh=20, connectivity=4
    )

    rasters_rio.write(sieve_arr, sieve_meta, sieve_out, nodata=255)
    ci.assert_raster_equal(sieve_out, raster_sieved_path)


@s3_env
def test_collocate(raster_meta):
    """Test collocate function"""
    raster, meta = raster_meta

    coll_arr, coll_meta = rasters_rio.collocate(
        meta, raster, meta
    )  # Just hope that it doesnt crash

    assert coll_meta == meta


@s3_env
def test_merge_gtiff(tmp_path, raster_path):
    """Test merge_gtiff function"""
    raster_to_merge_path = rasters_path().joinpath("raster_to_merge.tif")
    raster_merged_gtiff_path = rasters_path().joinpath("raster_merged.tif")
    raster_merged_gtiff_out = os.path.join(tmp_path, "test_merged.tif")

    rasters_rio.merge_gtiff(
        [raster_path, raster_to_merge_path],
        raster_merged_gtiff_out,
        method="max",
        driver="GTiff",
        dtype="uint8",
    )

    ci.assert_raster_equal(raster_merged_gtiff_out, raster_merged_gtiff_path)


@s3_env
def test_vectorize(raster_path):
    """Test vectorize function"""
    if shapely.__version__ >= "1.8a1":
        vect_truth_path = rasters_path().joinpath("vector.geojson")
        diss_truth_path = rasters_path().joinpath("dissolved.geojson")
    else:
        print("USING OLD VECTORS")
        vect_truth_path = rasters_path().joinpath("vector_old.geojson")
        diss_truth_path = rasters_path().joinpath("dissolved_old.geojson")

    # Vectorize
    val = 2
    vect = rasters_rio.vectorize(raster_path)
    vect_val = rasters_rio.vectorize(raster_path, values=val)
    vect_val_diss = rasters_rio.vectorize(raster_path, values=val, dissolve=True)
    vect_val_disc = rasters_rio.vectorize(
        raster_path, values=[1, 255], keep_values=False
    )
    vect_truth = vectors.read(vect_truth_path)
    diss_truth = vectors.read(diss_truth_path)
    ci.assert_geom_equal(vect, vect_truth)
    ci.assert_geom_equal(vect_val, vect_truth.loc[vect_truth.raster_val == val])
    ci.assert_geom_equal(vect_val_diss, diss_truth)
    ci.assert_geom_equal(vect_val_disc, vect_truth.loc[vect_truth.raster_val == val])


@s3_env
def test_get_valid_vector(raster_path):
    """Test get_valid_vector function"""
    valid_truth_path = rasters_path().joinpath("valid.geojson")

    valid_vec = rasters_rio.get_valid_vector(raster_path)

    valid_truth = vectors.read(valid_truth_path)
    ci.assert_geom_equal(valid_vec, valid_truth)


@s3_env
def test_nodata_vec(raster_path):
    """Test get_nodata_vector function"""
    nodata_truth_path = rasters_path().joinpath("nodata.geojson")

    nodata_vec = rasters_rio.get_nodata_vector(raster_path)

    nodata_truth = vectors.read(nodata_truth_path)
    ci.assert_geom_equal(nodata_vec, nodata_truth)


@s3_env
@pytest.mark.skipif(
    shutil.which("gdalbuildvrt") is None,
    reason="Only works if gdalbuildvrt can be found.",
)
def test_vrt(tmp_path, raster_path):
    raster_merged_vrt_path = rasters_path().joinpath("raster_merged.vrt")
    raster_to_merge_path = rasters_path().joinpath("raster_to_merge.tif")

    # Merge VRT
    raster_merged_vrt_out = os.path.join(tmp_path, "test_merged.vrt")
    rasters_rio.merge_vrt([raster_path, raster_to_merge_path], raster_merged_vrt_out)
    ci.assert_raster_equal(raster_merged_vrt_out, raster_merged_vrt_path)

    os.remove(raster_merged_vrt_out)

    rasters_rio.merge_vrt(
        [raster_path, raster_to_merge_path],
        raster_merged_vrt_out,
        abs_path=True,
        **KAPUT_KWARGS,
    )
    ci.assert_raster_equal(raster_merged_vrt_out, raster_merged_vrt_path)


def test_dim():
    """Test on BEAM-DIMAP function"""
    dim_path = rasters_path().joinpath("DIM.dim")
    dim_img_path = rasters_rio.get_dim_img_path(dim_path)
    assert dim_img_path.is_file(), f"{dim_img_path} is not a file!"
    assert dim_img_path == rasters_path().joinpath("DIM.data", "dim.img")


def test_bit():
    """Test bit arrays"""
    np_ones = np.ones((1, 2, 2), dtype=np.uint16)
    ones = rasters_rio.read_bit_array(np_ones, bit_id=0)
    zeros = rasters_rio.read_bit_array(np_ones, bit_id=list(np.arange(1, 15)))
    assert (np_ones == ones).all()
    for arr in zeros:
        assert (np_ones == 1 + arr).all()


@s3_env
def test_dem_fct(tmp_path):
    """Test DEM fct, i.e. slope and hillshade"""
    # Paths IN
    dem_path = rasters_path().joinpath("dem.tif")
    hlsd_path = rasters_path().joinpath("hillshade_rio.tif")
    slope_path = rasters_path().joinpath("slope_rio.tif")
    slope_r_path = rasters_path().joinpath("slope_r_rio.tif")
    slope_p_path = rasters_path().joinpath("slope_p_rio.tif")

    # Path OUT
    hlsd_path_out = os.path.join(tmp_path, "hillshade_out_rio.tif")
    slope_path_out = os.path.join(tmp_path, "slope_out_rio.tif")
    slope_r_path_out = os.path.join(tmp_path, "slope_out_r_rio.tif")
    slope_p_path_out = os.path.join(tmp_path, "slope_out_p_rio.tif")

    # Compute
    hlsd, meta = rasters_rio.hillshade(dem_path, 34.0, 45.2)
    rasters_rio.write(hlsd, meta, hlsd_path_out, dtype="float32")

    slp, meta = rasters_rio.slope(dem_path)
    rasters_rio.write(slp, meta, slope_path_out, dtype="float32")

    slp_r, meta = rasters_rio.slope(dem_path, in_pct=False, in_rad=True)
    rasters_rio.write(slp_r, meta, slope_r_path_out, dtype="float32")

    slp_p, meta = rasters_rio.slope(dem_path, in_pct=True)
    rasters_rio.write(slp_p, meta, slope_p_path_out, dtype="float32")

    # Test
    ci.assert_raster_almost_equal(hlsd_path, hlsd_path_out, decimal=4)
    ci.assert_raster_almost_equal(slope_path, slope_path_out, decimal=4)
    ci.assert_raster_almost_equal(slope_r_path, slope_r_path_out, decimal=4)
    ci.assert_raster_almost_equal(slope_p_path, slope_p_path_out, decimal=4)


@s3_env
def test_reproj(tmp_path, raster_path):
    """Test reproject fct"""
    dem_path = rasters_path().joinpath(
        "Copernicus_DSM_10_N43_00_W003_00_DEM_resampled.tif"
    )
    reproj_path = rasters_path().joinpath("reproj_out.tif")

    with rasterio.open(str(dem_path)) as src, rasterio.open(str(raster_path)) as ds:
        dst_arr, dst_meta = rasters_rio.reproject_match(
            ds.meta, src.read(), src.meta, **KAPUT_KWARGS
        )

        # from ds
        assert ds.meta["driver"] == dst_meta["driver"]
        assert ds.meta["width"] == dst_meta["width"]
        assert ds.meta["height"] == dst_meta["height"]
        assert ds.meta["transform"] == dst_meta["transform"]

        # from src
        assert src.meta["count"] == dst_meta["count"]
        assert src.meta["nodata"] == dst_meta["nodata"]
        assert src.meta["dtype"] == dst_meta["dtype"]

        path_out = os.path.join(tmp_path, "out.tif")
        rasters_rio.write(dst_arr, dst_meta, path_out)

        ci.assert_raster_almost_equal(path_out, reproj_path, decimal=4)


@s3_env
def test_rasterize(tmp_path, raster_path):
    """Test rasterize fct"""
    vec_path = rasters_path().joinpath("vector.geojson")
    raster_true_bin_path = rasters_path().joinpath("rasterized_bin.tif")
    raster_true_path = rasters_path().joinpath("rasterized.tif")

    # Binary vector
    out_bin_path = os.path.join(tmp_path, "out_bin.tif")
    rast_bin, bin_meta = rasters_rio.rasterize(raster_path, vec_path)
    rasters_rio.write(rast_bin, bin_meta, out_bin_path)

    ci.assert_raster_almost_equal(raster_true_bin_path, out_bin_path, decimal=4)

    # Vector
    out_path = os.path.join(tmp_path, "out.tif")
    rast, meta = rasters_rio.rasterize(
        raster_path, vec_path, "raster_val", dtype=np.uint8
    )
    rasters_rio.write(rast, meta, out_path)

    ci.assert_raster_almost_equal(raster_true_path, out_path, decimal=4)


def test_read_idx():
    """Test mtd after reading with index"""
    l1_path = rasters_path().joinpath("19760712T093233_L1_215030_MSS_stack.tif")

    def _test_idx(idx_list):
        raster, meta = rasters_rio.read(l1_path, indexes=idx_list)

        if isinstance(idx_list, list):
            nof_idx = len(idx_list)
            shape = (meta["count"], meta["height"], meta["width"])
        else:
            nof_idx = 1
            shape = (meta["height"], meta["width"])

        ci.assert_val(meta["count"], nof_idx, "Count")
        ci.assert_val(raster.shape, shape, "Shape")

    # Tests
    _test_idx([1])
    _test_idx([1, 2])
    _test_idx(1)


@s3_env
def test_decorator_deprecation(raster_path):
    from sertit.rasters_rio import path_arr_dst

    @any_raster_to_rio_ds
    def _ok_rasters(ds):
        return ds.read()

    @path_arr_dst
    def _depr_rasters(ds):
        return ds.read()

    # Not able to warn deprecation from inside the decorator
    np.testing.assert_equal(_ok_rasters(raster_path), _depr_rasters(raster_path))


def test_get_nodata_deprecation():
    """Test deprecation of get_nodata_value"""
    # Test deprecation
    for dtype in [
        np.uint8,
        np.int8,
        np.uint16,
        np.uint32,
        np.int32,
        np.int64,
        np.uint64,
        int,
        "int",
        np.int16,
        np.float32,
        np.float64,
        float,
        "float",
    ]:
        with pytest.deprecated_call():
            from sertit.rasters_rio import get_nodata_value

            ci.assert_val(
                get_nodata_value_from_dtype(dtype), get_nodata_value(dtype), dtype
            )


@s3_env
def test_write_deprecated(tmp_path, raster_path):
    test_deprecated_path = os.path.join(tmp_path, "test_depr.tif")
    raster, mtd = rasters_rio.read(raster_path)

    # test deprecation warning
    with pytest.deprecated_call():
        rasters_rio.write(raster, mtd, path=test_deprecated_path)



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing the files"""

import os
import shutil
import tempfile

import pytest

from ci.script_utils import files_path, get_s3_ci_path, s3_env
from sertit import AnyPath, ci, misc, path, vectors

ci.reduce_verbosity()


def test_paths():
    """Test path functions"""
    curr_file = AnyPath(__file__).resolve()
    curr_dir = curr_file.parent
    with misc.chdir(curr_dir):
        # Relative path
        curr_rel_path = path.real_rel_path(curr_file, curr_dir)
        assert curr_rel_path == AnyPath(os.path.join(".", os.path.basename(__file__)))

        # Abspath
        abs_file = path.to_abspath(curr_rel_path)
        assert abs_file == curr_file

        with pytest.raises(FileNotFoundError):
            path.to_abspath("haha.txt", raise_file_not_found=True)

        # with not pytest.raises(FileNotFoundError):
        path.to_abspath("haha.txt", raise_file_not_found=False)

        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp = path.to_abspath(os.path.join(tmp_dir, "haha"))
            assert os.path.isdir(tmp)

        # Listdir abspath
        list_abs = path.listdir_abspath(curr_dir)
        assert curr_file in list_abs

        # Root path
        assert str(abs_file).startswith(str(path.get_root_path()))

        # Writeable
        with tempfile.TemporaryDirectory() as tmp_dir:
            assert path.is_writable(tmp_dir)  # Writeable

        assert not path.is_writable(get_s3_ci_path())  # Not writable
        assert not path.is_writable("cvfgbherth")  # Non-existing


@s3_env
def test_archived_paths():
    landsat_name = "LM05_L1TP_200030_20121230_20200820_02_T2_CI"
    ok_folder = files_path().joinpath(landsat_name)
    zip_file = files_path().joinpath(f"{landsat_name}.zip")
    tar_file = files_path().joinpath(f"{landsat_name}.tar")
    targz_file = files_path().joinpath(f"{landsat_name}.tar.gz")
    sz_file = files_path().joinpath(f"{landsat_name}.7z")

    # Archive file
    tif_name = "LM05_L1TP_200030_20121230_20200820_02_T2_QA_RADSAT.TIF"
    tif_ok = f"{ok_folder.name}/{tif_name}"
    tif_regex = f".*{tif_name}"
    assert tif_ok == path.get_archived_path(zip_file, tif_regex)
    assert tif_ok == path.get_archived_path(zip_file, tif_regex, as_list=True)[0]
    assert tif_ok == path.get_archived_path(tar_file, ".*RADSAT")

    # RASTERIO
    tif_zip = path.get_archived_rio_path(zip_file, tif_regex)
    tif_list = path.get_archived_rio_path(zip_file, tif_regex, as_list=True)
    tif_tar = path.get_archived_rio_path(tar_file, ".*RADSAT")
    tif_ok = ok_folder.joinpath(tif_name)
    ci.assert_raster_equal(tif_ok, tif_zip)
    ci.assert_raster_equal(tif_ok, tif_list[0])
    ci.assert_raster_equal(tif_ok, tif_tar)

    file_list = path.get_archived_file_list(zip_file)
    ci.assert_raster_equal(
        tif_ok, path.get_archived_rio_path(zip_file, tif_regex, file_list=file_list)
    )

    # VECTORS
    vect_name = "map-overlay.kml"
    vec_ok_path = ok_folder.joinpath(vect_name)
    if shutil.which("ogr2ogr"):  # Only works if ogr2ogr can be found.
        vect_regex = f".*{vect_name}"
        vect_zip = vectors.read(zip_file, archive_regex=vect_regex)
        vect_tar = vectors.read(tar_file, archive_regex=r".*overlay\.kml")
        vect_ok = vectors.read(vec_ok_path)
        assert not vect_ok.empty
        ci.assert_geom_equal(vect_ok, vect_zip)
        ci.assert_geom_equal(vect_ok, vect_tar)

    # ERRORS
    with pytest.raises(TypeError):
        path.get_archived_rio_path(targz_file, tif_regex)
    with pytest.raises(TypeError):
        path.get_archived_rio_path(sz_file, tif_regex)
    with pytest.raises(FileNotFoundError):
        path.get_archived_rio_path(zip_file, "cdzeferf")


def test_get_file_name():
    """Test get_file_name"""
    file_name = path.get_filename(__file__)
    assert file_name == "test_path"
    file_name = path.get_filename(__file__ + "\\")
    assert file_name == "test_path"
    file_name = path.get_filename(__file__ + "/")
    assert file_name == "test_path"

    file = "/fkjzeh-r_éfertg.tar.gz"
    assert file[1:] == path.get_filename(file) + "." + path.get_ext(file)

    # Multi point files
    fn = r"/test/HLS.L30.T42RVR.2022240T055634.v2.0.B01.tif"
    file_name = path.get_filename(fn)
    assert file_name == "HLS.L30.T42RVR.2022240T055634.v2.0.B01"

    fn = (
        r"/test/S3B_SL_1_RBT____20200909T104016_0179_043_165_2340_LN2_O_NT_004.SEN3.zip"
    )
    file_name = path.get_filename(fn)
    assert file_name == "S3B_SL_1_RBT____20200909T104016_0179_043_165_2340_LN2_O_NT_004"

    fn = r"/test/S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
    file_name = path.get_filename(fn)
    assert file_name == "S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432"

    fn = r"/test/S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.tar.gz.zip"
    file_name = path.get_filename(fn, other_exts=".gz.zip")
    assert file_name == "S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432"


def test_find_files():
    """Test find_files"""
    names = os.path.basename(__file__)
    root_paths = AnyPath(__file__).parent
    max_nof_files = 1
    get_as_str = True

    # Test
    found_path = path.find_files(names, root_paths, max_nof_files, get_as_str)

    assert found_path == AnyPath(__file__)


def test_get_file_in_dir():
    """Test get_file_in_dir"""
    # Get parent dir
    folder = os.path.dirname(os.path.realpath(__file__))

    # Test
    file = path.get_file_in_dir(
        folder, "path", ".py", filename_only=False, get_list=True, exact_name=False
    )
    filename = path.get_file_in_dir(
        folder,
        path.get_filename(__file__),
        "py",
        filename_only=True,
        get_list=False,
        exact_name=True,
    )

    assert file[0] == AnyPath(__file__)
    assert filename == os.path.basename(__file__)



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing raster functions (with XARRAY)"""

import logging
import os
import shutil

import numpy as np
import pytest
import rasterio
import shapely
import xarray as xr

from ci.script_utils import (
    CI_SERTIT_USE_DASK,
    KAPUT_KWARGS,
    assert_chunked_computed,
    dask_env,
    get_output,
    is_not_lazy_yet,
    rasters_path,
    s3_env,
)
from sertit import ci, geometry, path, rasters, unistra, vectors
from sertit.rasters import (
    DEG_LAT_TO_M,
    FLOAT_NODATA,
    INT8_NODATA,
    UINT8_NODATA,
    UINT16_NODATA,
    any_raster_to_xr_ds,
    get_nodata_value_from_dtype,
    get_nodata_value_from_xr,
)
from sertit.vectors import EPSG_4326

ci.reduce_verbosity()

DEBUG = False


def test_indexes(caplog):
    @s3_env
    @dask_env(nof_computes=2)  # assert_equal x2
    def test_core():
        l1_path = rasters_path().joinpath("19760712T093233_L1_215030_MSS_stack.tif")
        xda_raw = rasters.read(l1_path, **KAPUT_KWARGS)
        xda_idx = rasters.read(l1_path, indexes=1)

        xr.testing.assert_equal(xda_raw[[0], :], xda_idx)  # Compute #1

        xda_idx2 = rasters.read(l1_path, indexes=[3, 2])
        xda_raw2 = np.concatenate((xda_raw.data[[2], :], xda_raw.data[[1], :]))

        with pytest.raises(ValueError):
            rasters.read(l1_path, indexes=0)

        with caplog.at_level(logging.WARNING):
            idx = [4, 5]
            rasters.read(l1_path, indexes=idx)
            assert f"Non available index: {idx}" in caplog.text

        # Test laziness
        assert_chunked_computed(xda_raw, "Read without index")
        assert_chunked_computed(xda_idx, "Read with one index")
        assert_chunked_computed(xda_idx2, "Read with indices")
        assert_chunked_computed(xda_raw2, "Concatenation")

        # Compute after checking the array is chunked (otherwise it fails!)
        import dask

        xda_raw2, xda_idx2 = dask.compute(xda_raw2, xda_idx2)  # Compute #2
        np.testing.assert_equal(xda_raw2, xda_idx2)

    test_core()


@pytest.fixture
def raster_path():
    return rasters_path().joinpath("raster.tif")


@pytest.fixture
def dem_path():
    return rasters_path().joinpath("dem.tif")


@pytest.fixture
def mask_path():
    return rasters_path().joinpath("raster_mask.geojson")


@pytest.fixture
def mask(mask_path):
    return vectors.read(mask_path)


@pytest.fixture
def ds_name(raster_path):
    with rasterio.open(str(raster_path)) as ds:
        return path.get_filename(ds.name)


@pytest.fixture
def ds_dtype(raster_path):
    with rasterio.open(str(raster_path)) as ds:
        return getattr(np, ds.meta["dtype"])


def get_xda(raster_path, **kwargs):
    return rasters.read(raster_path, **kwargs)


def get_xds(raster_path):
    with rasterio.open(str(raster_path)) as ds:
        ds_name = path.get_filename(ds.name)
    return xr.Dataset({ds_name: get_xda(raster_path)})


@s3_env
@dask_env()
def test_rasters_extent(tmp_path, raster_path, ds_name, ds_dtype):
    """"""
    extent_path = rasters_path().joinpath("extent.geojson")
    # Get Extent
    extent = rasters.get_extent(raster_path)
    truth_extent = vectors.read(extent_path)
    ci.assert_geom_equal(extent, truth_extent)


@s3_env
@dask_env(nof_computes=1)  # Vectorize is not lazy
def test_rasters_footprint(tmp_path, raster_path, ds_name, ds_dtype):
    # Get Footprint
    footprint_path = rasters_path().joinpath("footprint.geojson")
    footprint = rasters.get_footprint(raster_path)  # Compute #1
    truth_footprint = vectors.read(footprint_path)
    ci.assert_geom_equal(footprint, truth_footprint)


@s3_env
@dask_env(
    nof_computes=1
)  # TODO: Read with downsampling issue https://github.com/opendatacube/odc-geo/issues/236
def test_read(tmp_path, raster_path, ds_name, ds_dtype):
    """Test read function"""
    # xda, xds
    xda = get_xda(raster_path)
    xds = get_xds(raster_path)

    with rasterio.open(str(raster_path)) as ds:
        # -- Read and test laziness
        assert_chunked_computed(xda, "Native xda")

        # Native resolution
        xda_1 = rasters.read(ds, resolution=ds.res[0])
        assert_chunked_computed(xda_1, "Read with native resolution (x)")

        # Native resolution
        xda_2 = rasters.read(raster_path, resolution=[ds.res[0], ds.res[1]])
        assert_chunked_computed(xda_2, "Read with native resolution (x,y)")

        # Native size
        xda_3 = rasters.read(raster_path, size=(xda_1.rio.width, xda_1.rio.height))
        assert_chunked_computed(xda_3, "Read with native size")

        # Upsampling (reproject)
        xda_4 = rasters.read(raster_path, resolution=ds.res[0] / 2)
        assert_chunked_computed(xda_4, "Read with upsampling")

        # Read a xarray
        xda_5 = rasters.read(xda)
        assert_chunked_computed(xda_5, "Read an already existing xr.DataArray")

        # Downsampling (coarsen)
        xda_6 = rasters.read(raster_path, resolution=ds.res[0] * 2)
        assert_chunked_computed(xda_6, "Read with downsampling")

        # Downsampling (reproject)
        xda_7 = rasters.read(raster_path, resolution=ds.res[0] * 10.1)  # Compute #1
        # TODO: remove when https://github.com/opendatacube/odc-geo/issues/236 is fixed
        # assert_chunked_computed(xda_7, "Read with downsampling (reproject)")

        # Test shape (link between resolution and size)
        assert xda_4.shape[-2] == xda.shape[-2] * 2
        assert xda_4.shape[-1] == xda.shape[-1] * 2
        assert xda_6.shape[-2] == xda.shape[-2] / 2
        assert xda_6.shape[-1] == xda.shape[-1] / 2
        assert xda_7.shape[-1] == round(xda.shape[-1] / 10.1)
        assert xda_7.shape[-2] == round(xda.shape[-2] / 10.1)
        with pytest.raises(ValueError):
            rasters.read(ds, resolution=[20, 20, 20])

        # Test dataset integrity
        assert xda.shape == (ds.count, ds.height, ds.width)
        assert xda.encoding["dtype"] == ds_dtype
        assert xds[ds_name].shape == xda.shape
        assert xda_1.rio.crs == ds.crs
        assert xda_1.rio.transform() == ds.transform
        xr.testing.assert_equal(xda_1, xda_2)
        xr.testing.assert_equal(xda_1, xda_3)
        xr.testing.assert_equal(xda, xda_5)

        ci.assert_xr_encoding_attrs(xda, xda_1)
        ci.assert_xr_encoding_attrs(xda, xda_2)
        ci.assert_xr_encoding_attrs(xda, xda_3)
        ci.assert_xr_encoding_attrs(xda, xda_4)
        ci.assert_xr_encoding_attrs(xda, xda_5)
        ci.assert_xr_encoding_attrs(xda, xda_6)
        ci.assert_xr_encoding_attrs(xda, xda_7)
        ci.assert_val(xda_1.attrs["path"], str(raster_path), "raster path")


@s3_env
def test_read_np_vs_dask(raster_path):
    xda = get_xda(raster_path, chunks=None)

    # Should be always lazy
    xda_dask = rasters.read(raster_path, chunks=True)
    ci.assert_chunked(xda_dask), "Read with chunks = True"

    # Test arrays are the same
    xr.testing.assert_equal(xda, xda_dask)
    ci.assert_xr_encoding_attrs(xda, xda_dask, unchecked_attr="preferred_chunks")


@s3_env
@dask_env(nof_computes=2)  # Write x2
def test_read_with_window(tmp_path, raster_path, mask_path, mask):
    """Test read (with window) function"""
    raster_window_path = rasters_path().joinpath("window.tif")
    raster_window_20_path = rasters_path().joinpath("window_20.tif")

    # Window with native resolution
    xda_window_out = get_output(tmp_path, "test_xda_window.tif", DEBUG)
    xda_window = rasters.read(
        raster_path,
        window=mask_path,
    )
    assert_chunked_computed(xda_window, "Read with a window with native resolution")
    rasters.write(xda_window, xda_window_out, dtype=np.uint8)  # Compute #1
    ci.assert_raster_equal(xda_window_out, raster_window_path)

    # Window with updated resolution
    xda_window_20_out = get_output(tmp_path, "test_xda_20_window.tif", DEBUG)
    gdf = mask.to_crs(EPSG_4326)
    xda_window_20 = rasters.read(raster_path, window=gdf, resolution=20)
    ci.assert_val(round(xda_window_20.rio.resolution()[0]), 20, "resolution")

    # TODO: https://github.com/opendatacube/odc-geo/issues/236
    # assert_chunked_computed(xda_window_20, "Read with a window with updated resolution")

    rasters.write(xda_window_20, xda_window_20_out, dtype=np.uint8)  # Compute #2
    ci.assert_raster_equal(xda_window_20_out, raster_window_20_path)

    # Non existing window
    with pytest.raises(FileNotFoundError):
        rasters.read(
            raster_path,
            window=rasters_path().joinpath("non_existing_window.kml"),
        )


@s3_env
@dask_env()
@pytest.mark.timeout(4, func_only=True)
def test_very_big_file(tmp_path, mask_path):
    """Test read with big files function (should be fast, if not there is something going on...)"""
    dem_path = unistra.get_geodatastore() / "GLOBAL" / "EUDEM_v2" / "eudem_wgs84.tif"
    aoi_path = rasters_path() / "DAX.shp"
    with rasterio.open(dem_path) as ds:
        dem_res = ds.res[0]  # noqa
        dem_crs = ds.crs  # noqa

    aoi = vectors.read(aoi_path)
    cropped = rasters.crop(rasters.read(dem_path, window=aoi), aoi, nodata=0)
    assert_chunked_computed(cropped, "Crop (big file)")


@s3_env
@dask_env(max_computes=3)  # Write x2 or 3
@pytest.mark.parametrize(
    ("dtype", "nodata_val"),
    [
        pytest.param(np.uint8, UINT8_NODATA),
        pytest.param(np.int8, INT8_NODATA),
        pytest.param(np.uint16, UINT16_NODATA),
        pytest.param(np.int16, FLOAT_NODATA),
        pytest.param(np.uint32, UINT16_NODATA),
        pytest.param(np.int32, UINT16_NODATA),
        pytest.param(np.float32, FLOAT_NODATA),
        pytest.param(np.float64, FLOAT_NODATA),
    ],
)
def test_write(dtype, nodata_val, tmp_path, raster_path):
    """Test write (global check + cogs) function"""
    xda = get_xda(raster_path)
    dtype_str = dtype.__name__

    test_path = get_output(tmp_path, f"test_nodata_{dtype_str}.tif", DEBUG)
    test_cog_path = get_output(tmp_path, f"test_cog_{dtype_str}.tif", DEBUG)
    test_cog_no_dask_path = get_output(
        tmp_path, f"test_cog_no_dask{dtype_str}.tif", DEBUG
    )

    # Force negative value if possible
    if "uint" not in dtype_str:
        xda.data[:, 0, -1] = -3

    # -------------------------------------------------------------------------------------------------
    # Test GeoTiffs
    rasters.write(xda, test_path, dtype=dtype, **KAPUT_KWARGS)  # Compute #1
    _test_raster_after_write(test_path, dtype, nodata_val)

    # -------------------------------------------------------------------------------------------------
    # Test COGs
    if dtype not in [np.int8]:
        rasters.write(
            xda,
            test_cog_path,
            dtype=dtype,
            driver="COG",
            **KAPUT_KWARGS,
        )  # Compute #2
        _test_raster_after_write(test_path, dtype, nodata_val)

    # -------------------------------------------------------------------------------------------------
    # COGs without dask
    if dtype not in [np.int8]:
        rasters.write(
            xda,
            test_cog_no_dask_path,
            dtype=dtype,
            driver="COG",
            write_cogs_with_dask=False,
            **KAPUT_KWARGS,
        )  # Compute #3
        _test_raster_after_write(test_path, dtype, nodata_val)

    # test deprecation warning
    test_deprecated_path = get_output(tmp_path, "test_depr.tif", DEBUG)
    with pytest.deprecated_call():
        rasters.write(
            xda, path=test_deprecated_path, dtype=dtype
        )  # This doesn't compute (because of context manager ?)


@s3_env
# @dask_env(nof_compute=1)  # Write x1: TODO: Right now, Zarr doesn't seem to work with dask and GDAL 3.6.
def test_write_zarr(tmp_path, raster_path):
    # test zarr
    xda = get_xda(raster_path, chunks=None)
    zarr_path = get_output(tmp_path, "z.zarr", DEBUG)
    rasters.write(xda, path=zarr_path, driver="Zarr")
    xr.testing.assert_equal(xda, rasters.read(zarr_path, driver="Zarr"))


@s3_env
@dask_env(nof_computes=2)  # Write x2
def test_write_basic(tmp_path, raster_path, ds_dtype):
    """Test write (basic) function"""

    # xda, xds
    xda = get_xda(raster_path)
    xds = get_xds(raster_path)

    # DataArray
    xda_out = get_output(tmp_path, "test_xda.tif", DEBUG)
    rasters.write(xda, xda_out, dtype=ds_dtype)
    assert os.path.isfile(xda_out)

    # Dataset
    xds_out = get_output(tmp_path, "test_xds.tif", DEBUG)
    rasters.write(xds, xds_out, dtype=ds_dtype)
    assert os.path.isfile(xds_out)

    # Tests
    ci.assert_raster_equal(raster_path, xda_out)
    ci.assert_raster_equal(raster_path, xds_out)


@s3_env
@dask_env(nof_computes=1)  # Write x1
def test_write_png(tmp_path, raster_path):
    """Test write (PNG) function"""

    # Just test if this doesn't fail
    xda_out = get_output(tmp_path, "test_xda.png", DEBUG)
    rasters.write(get_xda(raster_path), xda_out, dtype="uint8", driver="PNG")


@s3_env
@dask_env(nof_computes=1)  # Write x2 with dask, but only one compute!
def test_write_dask(tmp_path, raster_path, ds_dtype):
    """Test write (basic) function"""
    # xda, xds
    xda = get_xda(raster_path)
    xds = get_xds(raster_path)

    # DataArray
    xda_out = get_output(tmp_path, "test_xda.tif", DEBUG)
    delayed_1 = rasters.write(xda, xda_out, dtype=ds_dtype, compute=False)

    # Dataset
    xds_out = get_output(tmp_path, "test_xds.tif", DEBUG)
    delayed_2 = rasters.write(xds, xds_out, dtype=ds_dtype, compute=False)

    # Compute
    import dask

    dask.compute(delayed_1, delayed_2)
    assert os.path.isfile(xda_out)
    assert os.path.isfile(xds_out)

    # Tests
    ci.assert_raster_equal(raster_path, xda_out)
    ci.assert_raster_equal(raster_path, xds_out)


@s3_env
@dask_env(nof_computes=2)  # Write x2
def test_mask(tmp_path, raster_path, mask):
    """Test mask function"""
    # xda, xds
    xda = get_xda(raster_path)
    xds = get_xds(raster_path)

    # DataArray
    xda_masked = get_output(tmp_path, "test_mask_xda.tif", DEBUG)
    mask_xda = rasters.mask(xda, mask.geometry, **KAPUT_KWARGS)
    assert_chunked_computed(mask_xda, "Mask DataArray")
    rasters.write(mask_xda, xda_masked, dtype=np.uint8)
    ci.assert_xr_encoding_attrs(xda, mask_xda)

    # Dataset
    xds_masked = get_output(tmp_path, "test_mask_xds.tif", DEBUG)
    mask_xds = rasters.mask(xds, mask)
    assert_chunked_computed(mask_xds, "Mask Dataset")
    rasters.write(mask_xds, xds_masked, dtype=np.uint8)
    ci.assert_xr_encoding_attrs(xds, mask_xds)

    raster_masked_path = rasters_path().joinpath("raster_masked.tif")
    ci.assert_raster_equal(xda_masked, raster_masked_path)
    ci.assert_raster_equal(xds_masked, raster_masked_path)


@s3_env
@dask_env(nof_computes=4)  # Write x4
def test_paint(tmp_path, raster_path, mask):
    """Test paint function"""
    # xda, xds
    xda = get_xda(raster_path)
    xds = get_xds(raster_path)

    # -- DataArray --
    # Invert = True
    xda_paint_true = get_output(tmp_path, "test_paint_true_xda.tif", DEBUG)
    paint_true_xda = rasters.paint(
        xda, mask.geometry, value=600, invert=True, **KAPUT_KWARGS
    )
    assert_chunked_computed(paint_true_xda, "Paint DataArray (invert = true)")
    rasters.write(paint_true_xda, xda_paint_true, dtype=np.uint8)
    ci.assert_xr_encoding_attrs(xda, paint_true_xda)

    # Invert = False
    xda_paint_false = get_output(tmp_path, "test_paint_false_xda.tif", DEBUG)
    paint_false_xda = rasters.paint(xda, mask.geometry, value=600, invert=False)
    assert_chunked_computed(paint_false_xda, "Paint DataArray (invert = false)")
    rasters.write(paint_false_xda, xda_paint_false, dtype=np.uint8)
    ci.assert_xr_encoding_attrs(xda, paint_false_xda)

    # -- Dataset --
    # Invert = True
    xds_paint_true = get_output(tmp_path, "test_paint_true_xds.tif", DEBUG)
    paint_true_xds = rasters.paint(xds, mask, value=600, invert=True)
    assert_chunked_computed(paint_true_xds, "Paint Dataset (invert = true)")
    rasters.write(paint_true_xds, xds_paint_true, dtype=np.uint8)
    ci.assert_xr_encoding_attrs(xds, paint_true_xds)

    # Invert = False
    xds_paint_false = get_output(tmp_path, "test_paint_false_xds.tif", DEBUG)
    paint_false_xds = rasters.paint(xds, mask, value=600, invert=False)
    assert_chunked_computed(paint_false_xds, "Paint Dataset (invert = false)")
    rasters.write(paint_false_xds, xds_paint_false, dtype=np.uint8)
    ci.assert_xr_encoding_attrs(xds, paint_false_xds)


@s3_env
@dask_env(nof_computes=2)  # Write x2
def test_crop(tmp_path, raster_path, mask):
    """Test crop function"""
    # xda, xds
    xda = get_xda(raster_path)
    xds = get_xds(raster_path)

    # DataArray
    xda_cropped = get_output(tmp_path, "test_crop_xda.tif", DEBUG)
    crop_xda = rasters.crop(xda, mask.geometry, **KAPUT_KWARGS)
    assert_chunked_computed(crop_xda, "Crop DataArray")
    rasters.write(crop_xda, xda_cropped, dtype=np.uint8)  # Compute #1
    ci.assert_xr_encoding_attrs(xda, crop_xda)

    # Dataset
    xds_cropped = get_output(tmp_path, "test_crop_xds.tif", DEBUG)
    crop_xds = rasters.crop(xds, mask, nodata=get_nodata_value_from_xr(xds))
    assert_chunked_computed(crop_xds, "Crop Dataset")
    rasters.write(crop_xds, xds_cropped, dtype=np.uint8)  # Compute #2
    ci.assert_xr_encoding_attrs(xds, crop_xds)

    raster_cropped_xarray_path = rasters_path().joinpath("raster_cropped_xarray.tif")
    ci.assert_raster_equal(xda_cropped, raster_cropped_xarray_path)
    ci.assert_raster_equal(xds_cropped, raster_cropped_xarray_path)

    # Test with mask with Z
    mask_z = geometry.force_3d(mask)
    crop_z = rasters.crop(xda, mask_z)
    assert_chunked_computed(crop_z, "Crop 3D")
    xr.testing.assert_equal(crop_xda, crop_z)
    ci.assert_xr_encoding_attrs(crop_xda, crop_z)


@s3_env
@is_not_lazy_yet
@dask_env(
    nof_computes=2
)  # Sieve x2 (write a not-chunked array, so no additional 'compute')
def test_sieve(tmp_path, raster_path):
    """Test sieve function"""
    # TODO: not lazy

    # xda
    xda = get_xda(raster_path)

    # Sieve DataArray
    xda_sieved = get_output(tmp_path, "test_sieved_xda.tif", DEBUG)
    sieve_xda = rasters.sieve(xda, sieve_thresh=20, connectivity=4)
    assert_chunked_computed(sieve_xda, "Sieve DataArray")
    rasters.write(sieve_xda, xda_sieved, dtype=np.uint8)
    ci.assert_xr_encoding_attrs(xda, sieve_xda)

    # Test equality
    raster_sieved_path = rasters_path().joinpath("raster_sieved.tif")
    ci.assert_raster_equal(xda_sieved, raster_sieved_path)

    # Sieve From path
    sieve_xda_path = rasters.sieve(raster_path, sieve_thresh=20, connectivity=4)
    assert_chunked_computed(sieve_xda_path, "Sieve DataArray (directly from path)")
    xr.testing.assert_equal(sieve_xda, sieve_xda_path)


@s3_env
@is_not_lazy_yet
@dask_env(nof_computes=2)  # Sieve x2
def test_sieve_different_dtypes(tmp_path, raster_path):
    # Test with different dtypes

    # xda
    xda = get_xda(raster_path)

    sieve_xda_float = rasters.sieve(
        xda.astype(np.uint8).astype(np.float32), sieve_thresh=20, connectivity=4
    )
    assert_chunked_computed(sieve_xda_float, "Sieve DataArray (float)")
    sieve_xda_uint = rasters.sieve(
        xda.astype(np.uint8), sieve_thresh=20, connectivity=4
    )
    assert_chunked_computed(sieve_xda_uint, "Sieve DataArray (uint)")
    xr.testing.assert_equal(sieve_xda_uint, sieve_xda_float)


@s3_env
@is_not_lazy_yet
@dask_env(
    nof_computes=1
)  # Sieve x1 (write a not-chunked array, so no additional 'compute')
def test_sieve_dataset(tmp_path, raster_path):
    # xds
    xds = get_xds(raster_path)

    # Dataset
    xds_sieved = get_output(tmp_path, "test_sieved_xds.tif", DEBUG)
    sieve_xds = rasters.sieve(xds, sieve_thresh=20, connectivity=4)
    assert_chunked_computed(sieve_xds, "Sieve Dataset")
    rasters.write(sieve_xds, xds_sieved, dtype=np.uint8)
    ci.assert_xr_encoding_attrs(xds, sieve_xds)

    # Tests
    raster_sieved_path = rasters_path().joinpath("raster_sieved.tif")
    ci.assert_raster_equal(xds_sieved, raster_sieved_path)


@s3_env
@dask_env()
def test_collocate_self(tmp_path, raster_path):
    """Test collocate (with itself) functions"""
    # xda, xds
    xda = get_xda(raster_path)
    xds = get_xds(raster_path)

    # DataArray
    coll_xda = rasters.collocate(xda, xda, **KAPUT_KWARGS)
    assert_chunked_computed(coll_xda, "Collocate DataArray (self)")
    xr.testing.assert_equal(coll_xda, xda)
    ci.assert_xr_encoding_attrs(xda, coll_xda)

    # Dataset
    coll_xds = rasters.collocate(xds, xds)
    assert_chunked_computed(coll_xds, "Collocate Dataset (self)")
    xr.testing.assert_equal(coll_xds, xds)
    ci.assert_xr_encoding_attrs(xds, coll_xds)

    # Dataset with dataarray
    coll_xds = rasters.collocate(reference=xda, other=xds)
    assert_chunked_computed(coll_xds, "Collocate Dataset with DataArray")
    xr.testing.assert_equal(coll_xds, xds)
    ci.assert_xr_encoding_attrs(xds, coll_xds)


@s3_env
@dask_env()
def test_collocate(tmp_path):
    """Test collocate functions"""

    def __test_collocate_output(ref, other, coll, dtype):
        """Test collocated outputs"""
        assert_chunked_computed(coll, f"Collocate DataArray ({dtype})")

        # Keeps the same attrs as the original array (attrs, encoding, dtype and name)
        ci.assert_xr_encoding_attrs(other, coll)
        ci.assert_val(coll.dtype, other.dtype, f"Collocated dtype ({dtype})")
        ci.assert_val(coll.name, other.name, f"Collocated name ({float})")

        # But located on the reference one
        assert ref.coords.identical(coll.coords)
        with pytest.raises(AssertionError):
            ci.assert_val(ref.coords, other.coords, f"Raw coordinates ({dtype})")

    # Inputs
    other_float = rasters.read(
        rasters_path().joinpath("20191115T233722_S3_SLSTR_RBT_CLOUDS_25000-00m.tif"),
        masked=True,
        as_type=np.float32,
    )
    other_uint8 = rasters.read(
        rasters_path().joinpath(
            "20191115T233722_S3_SLSTR_RBT_CLOUDS_25000-00m_uint8.tif"
        ),
        masked=False,
        as_type=np.uint8,
    )
    ref = rasters.read(
        rasters_path().joinpath("20191115T233722_S3_SLSTR_RBT_HILLSHADE_MERIT_DEM.tif")
    )

    # Other in float
    coll_float = rasters.collocate(reference=ref, other=other_float)
    __test_collocate_output(ref, other_float, coll_float, "float")

    # Other in uint8
    coll_uint8 = rasters.collocate(reference=ref, other=other_uint8)
    __test_collocate_output(ref, other_uint8, coll_uint8, "uint8")


@s3_env
def test_merge_gtiff(tmp_path, raster_path):
    """Test merge_gtiff function"""
    raster_to_merge_path = rasters_path().joinpath("raster_to_merge.tif")
    raster_merged_gtiff_out = get_output(tmp_path, "test_merged.tif", DEBUG)
    rasters.merge_gtiff(
        [raster_path, raster_to_merge_path],
        raster_merged_gtiff_out,
        method="max",
        driver="GTiff",
        dtype="uint8",
    )

    raster_merged_gtiff_path = rasters_path().joinpath("raster_merged.tif")
    ci.assert_raster_equal(raster_merged_gtiff_out, raster_merged_gtiff_path)


@s3_env
@dask_env(nof_computes=5)  # vectorize x5 (vectorize not lazy yet)
def test_vectorize(tmp_path, raster_path, ds_name):
    """Test vectorize function"""
    # xda, xds
    xda = get_xda(raster_path)
    xds = get_xds(raster_path)

    if shapely.__version__ >= "1.8a1":
        vect_truth_path = rasters_path().joinpath("vector.geojson")
        diss_truth_path = rasters_path().joinpath("dissolved.geojson")
    else:
        print("USING OLD VECTORS")
        vect_truth_path = rasters_path().joinpath("vector_old.geojson")
        diss_truth_path = rasters_path().joinpath("dissolved_old.geojson")

    val = 2
    vect_truth = vectors.read(vect_truth_path)

    # DataArray
    vect_xda = rasters.vectorize(xda)
    vect_val = rasters.vectorize(raster_path, values=val)
    vect_val_diss = rasters.vectorize(raster_path, values=val, dissolve=True)
    vect_val_disc = rasters.vectorize(raster_path, values=[1, 255], keep_values=False)
    ci.assert_geom_equal(vect_xda, vect_truth)
    ci.assert_geom_equal(vect_val_diss, diss_truth_path)
    ci.assert_geom_equal(vect_val, vect_truth.loc[vect_truth.raster_val == val])
    ci.assert_geom_equal(vect_val_disc, vect_truth.loc[vect_truth.raster_val == val])

    # Dataset
    vect_xds = rasters.vectorize(xds)
    ci.assert_geom_equal(vect_xds[ds_name], vect_truth)


@s3_env
@dask_env(nof_computes=2)  # vectorize x2 (vectorize not lazy yet)
def test_get_valid_vec(tmp_path, raster_path, ds_name):
    """Test get_valid_vector function"""
    # xda, xds
    xda = get_xda(raster_path)
    xds = get_xds(raster_path)

    valid_truth_path = rasters_path().joinpath("valid.geojson")

    # -- Get valid vec
    valid_truth = vectors.read(valid_truth_path)

    # DataArray
    valid_vec = rasters.get_valid_vector(xda)
    ci.assert_geom_equal(valid_vec, valid_truth)

    # Dataset
    valid_vec_xds = rasters.get_valid_vector(xds)
    ci.assert_geom_equal(valid_vec_xds[ds_name], valid_truth)


@s3_env
@dask_env(nof_computes=2)  # vectorize x2 (vectorize not lazy yet)
def test_nodata_vec(tmp_path, raster_path, ds_name):
    """Test get_nodata_vector function"""
    nodata_truth_path = rasters_path().joinpath("nodata.geojson")

    # xda, xds
    xda = get_xda(raster_path)
    xds = get_xds(raster_path)

    # -- Get nodata vec
    nodata_truth = vectors.read(nodata_truth_path)

    # DataArray
    nodata_vec = rasters.get_nodata_vector(xda)
    ci.assert_geom_equal(nodata_vec, nodata_truth)

    # Dataset
    nodata_vec_xds = rasters.get_nodata_vector(xds)
    ci.assert_geom_equal(nodata_vec_xds[ds_name], nodata_truth)


@s3_env
@pytest.mark.skipif(
    shutil.which("gdalbuildvrt") is None,
    reason="Only works if gdalbuildvrt can be found.",
)
def test_vrt(tmp_path, raster_path):
    """Test merge_vrt function"""
    # SAME CRS
    raster_merged_vrt_path = rasters_path().joinpath("raster_merged.vrt")
    raster_to_merge_path = rasters_path().joinpath("raster_to_merge.tif")
    raster_path = rasters_path().joinpath("raster.tif")

    # Merge VRT
    raster_merged_vrt_out = get_output(tmp_path, "test_merged.vrt", DEBUG)
    rasters.merge_vrt(
        [raster_path, raster_to_merge_path], raster_merged_vrt_out, **KAPUT_KWARGS
    )
    ci.assert_raster_equal(raster_merged_vrt_out, raster_merged_vrt_path)

    os.remove(raster_merged_vrt_out)

    rasters.merge_vrt(
        [raster_path, raster_to_merge_path], raster_merged_vrt_out, abs_path=True
    )
    ci.assert_raster_equal(raster_merged_vrt_out, raster_merged_vrt_path)


@s3_env
@pytest.mark.skipif(
    shutil.which("gdalbuildvrt") is None,
    reason="Only works if gdalbuildvrt can be found.",
)
def test_merge_different_crs_rel(tmp_path):
    """Test merge_vrt (with different CRS) function"""
    # DIFFERENT CRS
    true_vrt_path = rasters_path().joinpath("merge_32-31.vrt")

    raster_1_path = rasters_path().joinpath(
        "20220228T102849_S2_T31TGN_L2A_134712_RED.tif"
    )
    raster_2_path = rasters_path().joinpath(
        "20220228T102849_S2_T32TLT_L2A_134712_RED.tif"
    )

    # Merge VRT
    raster_merged_vrt_out = get_output(tmp_path, "test_merged.vrt", DEBUG)
    rasters.merge_vrt([raster_1_path, raster_2_path], raster_merged_vrt_out)
    ci.assert_raster_equal(raster_merged_vrt_out, true_vrt_path)


@s3_env
@pytest.mark.skipif(
    shutil.which("gdalbuildvrt") is None,
    reason="Only works if gdalbuildvrt can be found.",
)
def test_merge_different_crs_abs(tmp_path):
    """Test merge_vrt (with different CRS) function"""
    # DIFFERENT CRS
    true_vrt_path = rasters_path().joinpath("merge_32-31.vrt")

    raster_1_path = rasters_path().joinpath(
        "20220228T102849_S2_T31TGN_L2A_134712_RED.tif"
    )
    raster_2_path = rasters_path().joinpath(
        "20220228T102849_S2_T32TLT_L2A_134712_RED.tif"
    )

    raster_merged_vrt_out = get_output(tmp_path, "test_merged.vrt", DEBUG)
    rasters.merge_vrt(
        [raster_1_path, raster_2_path], raster_merged_vrt_out, abs_path=True
    )
    ci.assert_raster_equal(raster_merged_vrt_out, true_vrt_path)


@s3_env
def test_merge_different_crs_gtiff(tmp_path):
    """Test merge_vrt (with different CRS) function"""
    # DIFFERENT CRS
    true_tif_path = rasters_path().joinpath("merge_32-31.tif")

    raster_1_path = rasters_path().joinpath(
        "20220228T102849_S2_T31TGN_L2A_134712_RED.tif"
    )
    raster_2_path = rasters_path().joinpath(
        "20220228T102849_S2_T32TLT_L2A_134712_RED.tif"
    )
    # Merge GTiff
    raster_merged_tif_out = get_output(tmp_path, "test_merged.tif", DEBUG)
    rasters.merge_gtiff([raster_1_path, raster_2_path], raster_merged_tif_out)
    ci.assert_raster_max_mismatch(
        raster_merged_tif_out, true_tif_path, max_mismatch_pct=1e-4
    )


def _test_raster_after_write(test_path, dtype, nodata_val):
    with rasterio.open(str(test_path)) as ds:
        assert ds.meta["dtype"] == dtype or ds.meta["dtype"] == dtype.__name__
        assert ds.meta["nodata"] == nodata_val
        assert ds.read()[:, 0, 0] == nodata_val  # Check value

        # Test negative value
        if "uint" not in dtype.__name__:
            assert ds.read()[:, 0, -1] == -3


def test_dim():
    """Test on BEAM-DIMAP function"""
    dim_path = rasters_path().joinpath("DIM.dim")
    dim_img_path = rasters.get_dim_img_path(dim_path)
    assert dim_img_path.is_file(), f"{dim_img_path} is not a file!"
    assert dim_img_path == rasters_path().joinpath("DIM.data", "dim.img")


@s3_env
@dask_env(nof_computes=1)  # compute x1
@pytest.mark.parametrize(
    ("dtype", "bit_id"),
    [
        pytest.param(np.uint16, 15),
        pytest.param(np.uint8, 7),
        pytest.param(np.uint32, 31),
    ],
)
def test_read_bit_array(dtype, bit_id):
    """Test bit arrays"""
    data = np.ones((1, 2, 2), dtype=dtype)

    if os.environ[CI_SERTIT_USE_DASK] == "1":
        # If we want to use dask, convert this array to a dask array
        from dask.array import from_array

        data = from_array(data)
    # Bit
    np_ones = xr.DataArray(data)
    ones = rasters.read_bit_array(np_ones, bit_id=0)
    zeros = rasters.read_bit_array(np_ones, bit_id=list(np.arange(1, bit_id)))

    # Outputs an array (either dask or numpy)
    if os.environ[CI_SERTIT_USE_DASK] == "1":
        import dask

        delayed = []
        delayed.append(dask.delayed(np.testing.assert_array_equal)(np_ones.data, ones))
        for arr in zeros:
            delayed.append(
                dask.delayed(np.testing.assert_array_equal)(np_ones.data, 1 + arr)
            )

        dask.compute(delayed)
    else:
        np.testing.assert_array_equal(np_ones.data, ones)
        for arr in zeros:
            np.testing.assert_array_equal(np_ones.data, 1 + arr)


@s3_env
@dask_env(nof_computes=1)  # compute x1
@pytest.mark.parametrize(
    ("dtype", "bit_id"),
    [
        pytest.param(np.uint8, 7),
        pytest.param(float, 7),
    ],
)
def test_read_uint8_array(dtype, bit_id):
    # uint8
    data = np.ones((1, 2, 2), dtype=dtype)

    if os.environ[CI_SERTIT_USE_DASK] == "1":
        # If we want to use dask, convert this array to a dask array
        from dask.array import from_array

        data = from_array(data)

    np_ones = xr.DataArray(data)
    ones = rasters.read_uint8_array(np_ones, bit_id=0)
    zeros = rasters.read_uint8_array(np_ones, bit_id=list(np.arange(1, bit_id)))

    # Outputs an array (either dask or numpy)
    if os.environ[CI_SERTIT_USE_DASK] == "1":
        import dask

        delayed = []
        delayed.append(dask.delayed(np.testing.assert_array_equal)(np_ones.data, ones))
        for arr in zeros:
            delayed.append(
                dask.delayed(np.testing.assert_array_equal)(np_ones.data, 1 + arr)
            )

        dask.compute(delayed)
    else:
        np.testing.assert_array_equal(np_ones.data, ones)
        for arr in zeros:
            np.testing.assert_array_equal(np_ones.data, 1 + arr)


@s3_env
@dask_env(nof_computes=1)  # assert_equal x1
def test_set_nodata():
    """Test xarray functions"""
    nodata_val = 0
    data = [[1, nodata_val, nodata_val], [nodata_val, nodata_val, nodata_val]]
    if os.environ[CI_SERTIT_USE_DASK] == "1":
        # If we want to use dask, convert this array to a dask array
        from dask.array import from_array

        data = from_array(data)

    # Set nodata
    xda = xr.DataArray(
        dims=("x", "y"),
        data=data,
    )
    xda.rio.write_nodata(-9999, inplace=True, encoded=True)
    assert_chunked_computed(xda, "Write nodata (rioxarray)")

    xda_nodata = rasters.set_nodata(xda, nodata_val)
    assert_chunked_computed(xda_nodata, "Set nodata")

    nodata = xr.DataArray(
        dims=("x", "y"), data=[[1, np.nan, np.nan], [np.nan, np.nan, np.nan]]
    )
    xr.testing.assert_equal(xda_nodata, nodata)
    ci.assert_val(xda_nodata.rio.encoded_nodata, nodata_val, "Encoded nodata")
    ci.assert_val(xda_nodata.rio.nodata, np.nan, "Array nodata")


@s3_env
@dask_env()
def test_xarray_fct(raster_path):
    """Test xarray functions"""
    xda = get_xda(raster_path)
    xda_sum = xda + xda
    assert_chunked_computed(xda_sum, "Sum between two DataArrays")

    # Mtd
    xda_sum = rasters.set_metadata(xda_sum, xda, "sum")
    assert_chunked_computed(xda_sum, "Set metadata")

    ci.assert_val(xda_sum.rio.crs, xda.rio.crs, "CRS")
    assert np.isnan(xda_sum.rio.nodata)
    ci.assert_val(
        get_nodata_value_from_xr(xda_sum), get_nodata_value_from_xr(xda), "nodata"
    )

    ci.assert_val(xda_sum.attrs.pop("long_name"), "sum", "long name")
    ci.assert_val(xda_sum.attrs, xda.attrs, "attributes")
    ci.assert_val(xda_sum.encoding, xda.encoding, "encoding")
    ci.assert_val(xda_sum.rio.transform(), xda.rio.transform(), "transform")
    ci.assert_val(xda_sum.rio.width, xda.rio.width, "width")
    ci.assert_val(xda_sum.rio.height, xda.rio.height, "height")
    ci.assert_val(xda_sum.rio.count, xda.rio.count, "count")
    ci.assert_val(xda_sum.name, "sum", "name")


@dask_env(nof_computes=2)  # assert_equal x2
def test_where():
    """Test overloading of xr.where function"""
    new_name = "mask_A"

    data = [[1, 0, 5], [np.nan, 0, 0]]
    if os.environ[CI_SERTIT_USE_DASK] == "1":
        # If we want to use dask, convert this array to a dask array
        from dask.array import from_array

        data = from_array(data)

    xarr = xr.DataArray(dims=("x", "y"), data=data)
    mask_xarr = rasters.where(xarr > 3, 0, 1, xarr, new_name=new_name)
    assert_chunked_computed(mask_xarr, "Where")

    xr.testing.assert_equal(np.isnan(xarr), np.isnan(mask_xarr))

    assert mask_xarr.attrs.pop("long_name") == new_name
    assert xarr.attrs == mask_xarr.attrs
    xr.testing.assert_equal(
        mask_xarr,
        xr.DataArray(
            dims=("x", "y"), data=np.array([[1.0, 1.0, 0.0], [np.nan, 1.0, 1.0]])
        ),
    )


@s3_env
@dask_env(nof_computes=1)  # Write x1
def test_aspect(tmp_path, dem_path):
    """Test aspect function"""
    aspect_path = rasters_path().joinpath("aspect.tif")

    # Path OUT
    aspect_path_out = get_output(tmp_path, "aspect_out.tif", DEBUG)

    # Aspect
    aspect = rasters.aspect(dem_path)
    assert_chunked_computed(aspect, "Aspect")
    rasters.write(aspect, aspect_path_out, dtype="float32")
    ci.assert_raster_almost_equal(aspect_path, aspect_path_out, decimal=4)


@s3_env
@dask_env(nof_computes=1)  # Write x1
def test_hillshade(tmp_path, dem_path):
    """Test hillshade function"""
    hlsd_path = rasters_path().joinpath("hillshade.tif")
    hlsd_path_out = get_output(tmp_path, "hillshade_out.tif", DEBUG)

    # Hillshade
    hlsd = rasters.hillshade(dem_path, 34.0, 45.2)
    assert_chunked_computed(hlsd, "Hillshade")
    rasters.write(hlsd, hlsd_path_out, dtype="float32")
    ci.assert_raster_almost_equal(hlsd_path, hlsd_path_out, decimal=4)


@s3_env
@dask_env(nof_computes=1)  # Write x1
def test_slope(tmp_path, dem_path):
    """Test slope function"""
    slope_path = rasters_path().joinpath("slope.tif")
    slope_path_out = get_output(tmp_path, "slope_out.tif", DEBUG)
    # Slope
    slp = rasters.slope(dem_path)
    assert_chunked_computed(slp, "Slope")
    rasters.write(slp, slope_path_out, dtype="float32")
    ci.assert_raster_almost_equal(slope_path, slope_path_out, decimal=4)


@s3_env
@dask_env(nof_computes=1)  # Write x1
def test_slope_rad(tmp_path, dem_path):
    """Test slope (radian) function"""
    slope_r_path = rasters_path().joinpath("slope_r.tif")
    slope_r_path_out = get_output(tmp_path, "slope_r_out.tif", DEBUG)

    # Slope rad
    slp_r = rasters.slope(dem_path, in_pct=False, in_rad=True)
    assert_chunked_computed(slp_r, "Slope (radian)")
    rasters.write(slp_r, slope_r_path_out, dtype="float32")
    ci.assert_raster_almost_equal(slope_r_path, slope_r_path_out, decimal=4)


@s3_env
@dask_env(nof_computes=1)  # Write x1
def test_slope_pct(tmp_path, dem_path):
    """Test slope (pct) function"""
    slope_p_path = rasters_path().joinpath("slope_p.tif")
    slope_p_path_out = get_output(tmp_path, "slope_p_out.tif", DEBUG)

    # Slope pct
    slp_p = rasters.slope(dem_path, in_pct=True)
    assert_chunked_computed(slp_p, "Slope (%)")
    rasters.write(slp_p, slope_p_path_out, dtype="float32")
    ci.assert_raster_almost_equal(slope_p_path, slope_p_path_out, decimal=4)


@s3_env
@dask_env(nof_computes=1)  # Write x1
def test_rasterize_binary(tmp_path, raster_path):
    """Test rasterize fct"""
    xda = get_xda(raster_path)

    # Binary vector
    vec_path = rasters_path().joinpath("vector.geojson")
    raster_true_bin_path = rasters_path().joinpath("rasterized_bin.tif")
    out_bin_path = get_output(tmp_path, "out_bin.tif", DEBUG)

    # Rasterize
    rast_bin = rasters.rasterize(xda, vec_path, **KAPUT_KWARGS)
    assert_chunked_computed(rast_bin, "Rasterize DataArray (binary vector)")
    rasters.write(rast_bin, out_bin_path, dtype=np.uint8, nodata=255)

    ci.assert_raster_almost_equal(raster_true_bin_path, out_bin_path, decimal=4)


@s3_env
@dask_env(nof_computes=1)  # Write x1
def test_rasterize_float(tmp_path, raster_path):
    # Binary vector with floating point raster
    vec_path = rasters_path().joinpath("vector.geojson")
    raster_float_path = rasters_path().joinpath("raster_float.tif")
    raster_true_bin_path = rasters_path().joinpath("rasterized_bin.tif")
    out_bin_path = get_output(tmp_path, "out_bin_float.tif", DEBUG)

    # Rasterize
    rast_bin = rasters.rasterize(rasters.read(raster_float_path), vec_path)
    assert_chunked_computed(
        rast_bin, "Rasterize DataArray (binary vector with floating point raster)"
    )
    rasters.write(rast_bin, out_bin_path, dtype=np.uint8, nodata=255)

    ci.assert_raster_almost_equal(raster_true_bin_path, out_bin_path, decimal=4)


@s3_env
@dask_env(nof_computes=1)  # Write x1
def test_rasterize_value_field(tmp_path, raster_path):
    # Use value_field
    vec_path = rasters_path().joinpath("vector.geojson")
    raster_true_path = rasters_path().joinpath("rasterized.tif")
    out_path = get_output(tmp_path, "out_rasterized.tif", DEBUG)

    # Rasterize
    rast = rasters.rasterize(
        raster_path, vec_path, value_field="raster_val", dtype=np.uint8
    )
    assert_chunked_computed(rast, "Rasterize DataArray (use value_field)")
    rasters.write(rast, out_path, dtype=np.uint8, nodata=255)

    ci.assert_raster_almost_equal(raster_true_path, out_path, decimal=4)


@s3_env
@dask_env(nof_computes=1)  # Write x1
def test_rasterize_multi_band_raster(tmp_path, raster_path):
    xda = get_xda(raster_path)
    vec_path = rasters_path().joinpath("vector.geojson")
    raster_true_bin_path = rasters_path().joinpath("rasterized_bin.tif")

    # Multiband raster
    out_bin_mb_path = get_output(tmp_path, "out_bin_mb.tif", DEBUG)

    # Rasterize
    rast_bin = rasters.rasterize(
        xr.concat([xda, xda], dim="band"), vec_path, **KAPUT_KWARGS
    )
    assert_chunked_computed(rast_bin, "Rasterize DataArray (multiband raster)")
    rasters.write(rast_bin, out_bin_mb_path, dtype=np.uint8, nodata=255)

    ci.assert_raster_almost_equal(raster_true_bin_path, out_bin_mb_path, decimal=4)


@s3_env
def test_decorator_deprecation(raster_path):
    from sertit.rasters import path_xarr_dst

    @any_raster_to_xr_ds
    def _ok_rasters(xds):
        assert isinstance(xds, xr.DataArray)
        return xds

    @path_xarr_dst
    def _depr_rasters(xds):
        assert isinstance(xds, xr.DataArray)
        return xds

    # Not able to warn deprecation from inside the decorator
    xr.testing.assert_equal(_ok_rasters(raster_path), _depr_rasters(raster_path))


def test_get_nodata_deprecation():
    """Test deprecation of get_nodata_value"""
    # Test deprecation
    for dtype in [
        np.uint8,
        np.int8,
        np.uint16,
        np.uint32,
        np.int32,
        np.int64,
        np.uint64,
        int,
        "int",
        np.int16,
        np.float32,
        np.float64,
        float,
        "float",
    ]:
        with pytest.deprecated_call():
            from sertit.rasters import get_nodata_value

            ci.assert_val(
                get_nodata_value_from_dtype(dtype), get_nodata_value(dtype), dtype
            )


@s3_env
@dask_env()
def test_get_notata_from_xr(raster_path):
    """Test get_nodata_value_from_xr"""
    ci.assert_val(get_nodata_value_from_xr(rasters.read(raster_path)), 255, "nodata")

    raster_path = rasters_path().joinpath(
        "20220228T102849_S2_T31TGN_L2A_134712_RED.tif"
    )
    ci.assert_val(get_nodata_value_from_xr(rasters.read(raster_path)), 65535, "nodata")

    raster_path = rasters_path().joinpath(
        "Copernicus_DSM_10_N43_00_W003_00_DEM_resampled.tif"
    )
    ci.assert_val(get_nodata_value_from_xr(rasters.read(raster_path)), None, "nodata")

    raster_path = rasters_path().joinpath("dem.tif")
    ci.assert_val(get_nodata_value_from_xr(rasters.read(raster_path)), -9999, "nodata")


def test_deg_meters_conversion():
    """
    Test conversions between deg and meters
    https://wiki.openstreetmap.org/wiki/Precision_of_coordinates#Precision_of_latitudes
    """
    # from_deg_to_meters
    ci.assert_val(
        rasters.from_deg_to_meters(1), DEG_LAT_TO_M, "1 degree of latitude in meters"
    )
    ci.assert_val(
        rasters.from_deg_to_meters(1, lat=45),
        78573.71,
        "1 degree of longitude at 45 degrees of latitude in meters",
    )
    ci.assert_val(
        rasters.from_deg_to_meters(1, lat=45, decimals=0),
        78574,
        "1 degree of longitude at 45 degrees of latitude in meters with 0 decimal",
    )
    ci.assert_val(
        rasters.from_deg_to_meters(1, lat=45, decimals=0, average_lat_lon=True),
        round((DEG_LAT_TO_M + 78574) / 2),
        "1 degree of longitude at 45 degrees of latitude in meters (averaged) with 0 decimal",
    )

    # from_deg_to_meters
    ci.assert_val(
        rasters.from_meters_to_deg(0.5), 4.5e-06, "0.5 meter in degrees of latitude"
    )
    ci.assert_val(
        rasters.from_meters_to_deg(0.5, lat=45),
        6.4e-06,
        "0.5 meter in degree of longitude at 45 degrees of latitude",
    )
    ci.assert_val(
        rasters.from_meters_to_deg(0.5, lat=45, decimals=9),
        6.363e-06,
        "0.5 meter in degree of longitude at 45 degrees of latitude with 9 decimal",
    )
    ci.assert_val(
        rasters.from_meters_to_deg(0.5, lat=45, decimals=9, average_lat_lon=True),
        5.432e-06,
        "0.5 meter in degree of longitude at 45 degrees of latitude (averaged) with 9 decimal",
    )


@s3_env
@dask_env(nof_computes=1)  # write x1
def test_classify(tmp_path):
    """Test classify"""
    d_ndvi_path = rasters_path() / "20200824_S2_20200908_S2_dNDVI.tif"
    sev_truth = rasters_path() / "fire_sev_ndvi_truth.tif"
    sev_out = get_output(tmp_path, "fire_sev_ndvi.tif", DEBUG, dask_folders=True)
    sev = rasters.classify(
        rasters.read(d_ndvi_path), bins=[0.2, 0.55], values=[2, 3, 4]
    )
    assert_chunked_computed(sev, "Classify DataArray")
    rasters.write(sev, sev_out, dtype=np.uint8, nodata=255)
    ci.assert_raster_equal(sev_truth, sev_out)


@s3_env
# @dask_env(nof_computes=1)  # write x1 RPC is not lazy yet, no need to test it multiple times
def test_reproject_rpc(tmp_path):
    """Test reproject with rpc"""
    spot_path = (
        rasters_path()
        / "IMG_SPOT7_MS_001_A"
        / "DIM_SPOT7_MS_201602150257025_SEN_1671661101.XML"
    )
    copdem_path = rasters_path() / "Copernicus_DSM_10_S07_00_E106_00_DEM.tif"
    with rasterio.open(str(spot_path)) as ds:
        rpc_reproj = rasters.reproject(
            spot_path, rpcs=ds.rpcs, dem_path=copdem_path, dst_crs="epsg:4326", nodata=0
        )

    ci.assert_val(rpc_reproj.rio.crs.to_epsg(), 4326, "Reprojected CRS")
    ci.assert_val(rpc_reproj.rio.count, 4, "Reprojected count")
    ci.assert_val(rpc_reproj.rio.encoded_nodata, 0, "Reprojected encoded nodata")
    ci.assert_val(rpc_reproj.rio.nodata, np.nan, "Reprojected nodata")


@s3_env
@dask_env(nof_computes=0)
def test_reproject(tmp_path, raster_path):
    """Test reproject"""
    # xda
    xda = get_xda(raster_path)

    # Reproject to a pixel_size
    size_arr = rasters.reproject(xda, pixel_size=60)
    ci.assert_val(round(size_arr.rio.resolution()[0]), 60, "Reprojected pixel size")
    ci.assert_val(size_arr.rio.count, 1, "Reprojected count")
    with pytest.raises(AssertionError):
        ci.assert_val(round(xda.rio.resolution()[0]), 60, "Native pixel pize")

    # Reproject to a shape
    shape_arr = rasters.reproject(xda, shape=(161, 232))
    ci.assert_val(shape_arr.shape, (1, 161, 232), "Reprojected shape")
    ci.assert_val(shape_arr.rio.shape, (161, 232), "Reprojected shape (rio)")
    ci.assert_val(shape_arr.rio.count, 1, "Reprojected count")
    with pytest.raises(AssertionError):
        ci.assert_val(xda.rio.shape, (1, 161, 232), "Native shape")

    # Reproject to espg:4326
    wgs84_arr = rasters.reproject(xda, dst_crs="EPSG:4326")
    ci.assert_val(wgs84_arr.rio.crs.to_epsg(), 4326, "Reprojected CRS")
    with pytest.raises(AssertionError):
        ci.assert_val(xda.rio.crs.to_epsg(), 4326, "Native CRS")



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing the CI"""

import os
import tempfile

import pytest
import rasterio
from cloudpathlib import AnyPath, S3Client
from tempenv import tempenv

from ci.script_utils import CI_SERTIT_S3
from sertit import ci, misc, rasters, s3
from sertit.ci import AWS_ACCESS_KEY_ID, AWS_S3_ENDPOINT, AWS_SECRET_ACCESS_KEY
from sertit.unistra import (
    _get_db_path,
    get_db2_path,
    get_db3_path,
    get_db4_path,
    get_geodatastore,
    s3_env,
)

ci.reduce_verbosity()


def base_fct(value):
    raster_path = AnyPath("s3://sertit-sertit-utils-ci").joinpath(
        "DATA", "rasters", "raster.tif"
    )
    assert raster_path.client.client.meta.endpoint_url == "https://s3.unistra.fr"
    assert raster_path.is_file()
    assert rasters.read(raster_path).rio.count == 1


@s3_env
def with_s3():
    base_fct(1)
    return 1


def without_s3():
    S3Client().set_as_default_client()
    base_fct(None)


@s3_env
def test_unistra_s3():
    with tempenv.TemporaryEnvironment(
        {s3.USE_S3_STORAGE: "1", AWS_S3_ENDPOINT: None, CI_SERTIT_S3: "1"}
    ):
        # Test s3_env and define_s3_client (called inside)
        raster_path = AnyPath("s3://sertit-sertit-utils-ci").joinpath(
            "DATA", "rasters", "raster.tif"
        )
        assert raster_path.client.client.meta.endpoint_url == "https://s3.unistra.fr"
        assert raster_path.is_file()
        assert rasterio.open(str(raster_path)).count == 1
        assert rasters.read(raster_path).rio.count == 1

    # Test profile
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        credentials = (
            f"[unistra]\n"
            f"aws_access_key_id={os.getenv(AWS_ACCESS_KEY_ID)}\n"
            f"aws_secret_access_key={os.getenv(AWS_SECRET_ACCESS_KEY)}\n"
        )
        f.write(credentials)
        filename = f.name
    with tempenv.TemporaryEnvironment(
        {
            s3.USE_S3_STORAGE: "1",
            AWS_ACCESS_KEY_ID: None,
            AWS_SECRET_ACCESS_KEY: None,
            AWS_S3_ENDPOINT: None,
            "AWS_SHARED_CREDENTIALS_FILE": filename,
            CI_SERTIT_S3: "1",
        }
    ):
        # Test s3_env and define_s3_client (called inside)
        raster_path = AnyPath("s3://sertit-sertit-utils-ci").joinpath(
            "DATA", "rasters", "raster.tif"
        )
        assert raster_path.client.client.meta.endpoint_url == "https://s3.unistra.fr"
        assert raster_path.is_file()
        assert rasterio.open(str(raster_path)).count == 1
        assert rasters.read(raster_path).rio.count == 1
    # Test get_geodatastore without s3
    with tempenv.TemporaryEnvironment({s3.USE_S3_STORAGE: "0"}):
        assert str(get_geodatastore()).endswith("BASES_DE_DONNEES")


@pytest.mark.skipif(not misc.in_docker(), reason="Only works in docker")
def test_mnt():
    """Test mounted directories"""
    try:
        assert get_db2_path() == "/mnt/ds2_db2"
        assert get_db3_path() == "/mnt/ds2_db3"
        assert get_db4_path() == "/mnt/ds2_db4"
    except NotADirectoryError:
        pass

    with pytest.raises(NotADirectoryError):
        _get_db_path(5)



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing vector functions"""

import logging
import os
import tempfile
import warnings

import geopandas as gpd
import pytest
from rasterio import CRS
from shapely import wkt

from ci.script_utils import KAPUT_KWARGS, files_path, s3_env, vectors_path
from sertit import ci, files, path, vectors
from sertit.logs import SU_NAME
from sertit.vectors import EPSG_4326, DataSourceError

LOGGER = logging.getLogger(SU_NAME)

ci.reduce_verbosity()


def _assert_attributes(vec: gpd.GeoDataFrame, vec_path):
    ci.assert_val(vec.attrs["path"], str(vec_path), "path")
    ci.assert_val(vec.attrs["name"], path.get_filename(vec_path), "name")


@s3_env
def test_vectors():
    """Test geo functions"""
    shp_path = vectors_path().joinpath("aoi.shp")
    kml_path = vectors_path().joinpath("aoi.kml")
    wkt_path = vectors_path().joinpath("aoi.wkt")
    utm_path = vectors_path().joinpath("aoi.geojson")
    ci.assert_geom_equal(shp_path, utm_path)  # Test shp

    # Test 3D vectors
    # with pytest.raises(AssertionError):
    #     ci.assert_geom_equal(shp_path, utm_path, ignore_z=False)

    # KML
    vectors.set_kml_driver()  # An error will occur afterwards if this fails (we are attempting to open a KML file)

    # KML to WKT
    aoi_str_test = vectors.get_aoi_wkt(kml_path, as_str=True)
    aoi_str = (
        "POLYGON Z ((46.1947755465253067 32.4973553439109324 0.0000000000000000, "
        "45.0353174370802520 32.4976496856158974 0.0000000000000000, "
        "45.0355748149750283 34.1139970085580018 0.0000000000000000, "
        "46.1956059695554089 34.1144793800670882 0.0000000000000000, "
        "46.1947755465253067 32.4973553439109324 0.0000000000000000))"
    )
    assert aoi_str == aoi_str_test

    aoi = vectors.get_aoi_wkt(kml_path, as_str=False)

    # WKT to WKT
    aoi2 = vectors.get_aoi_wkt(wkt_path, as_str=False)

    # UTM to WKT
    aoi3 = vectors.get_aoi_wkt(utm_path, as_str=False)

    assert aoi.equals(aoi2)  # No reprojection, should be equal
    assert aoi.equals_exact(
        aoi3, tolerance=0.5 * 10**6
    )  # Reprojection, so almost equal
    assert wkt.dumps(aoi) == aoi_str

    # UTM and bounds
    aoi = vectors.read(kml_path, **KAPUT_KWARGS)
    _assert_attributes(aoi, kml_path)

    with pytest.deprecated_call():
        assert (
            vectors.corresponding_utm_projection(aoi.centroid.x, aoi.centroid.y)
            == "EPSG:32638"
        )
        assert CRS.from_string("EPSG:32638") == vectors.to_utm_crs(
            aoi.centroid.x, aoi.centroid.y
        )

    env = aoi.envelope[0]

    # Test kwargs (should be slightly not equal toi AOI to prove bbox does sth)
    with pytest.raises(AssertionError):
        ci.assert_geom_equal(vectors.read(kml_path, bbox=env).geometry, aoi.envelope)

    # GeoDataFrame
    geodf = vectors.get_geodf(env, aoi.crs)  # GeoDataFrame from Polygon
    ci.assert_geom_equal(geodf.geometry, aoi.envelope)
    ci.assert_geom_equal(
        vectors.get_geodf(geodf.geometry, aoi.crs), geodf
    )  # GeoDataFrame from Geoseries
    ci.assert_geom_equal(
        vectors.get_geodf([env], aoi.crs), geodf
    )  # GeoDataFrame from list of poly

    with pytest.raises(TypeError):
        vectors.get_geodf([1, 2, 3, 4, 5], aoi.crs)
    with pytest.raises(TypeError):
        vectors.get_geodf([1, 2], aoi.crs)


@s3_env
def test_kml():
    """Test KML files"""
    LOGGER.debug("Open GEARTH_POLY.kml")
    # Just check there is no issue when opening this file
    kml_path = vectors_path().joinpath("GEARTH_POLY.kml")
    kml = vectors.read(kml_path)
    _assert_attributes(kml, kml_path)
    assert not kml.empty

    # Check equivalence between two vector types (complex vector)
    LOGGER.debug("Open EMSR680_AOI03_DEL_PRODUCT_observedEventA_v1.kml")
    kml_path = vectors_path().joinpath(
        "EMSR680_AOI03_DEL_PRODUCT_observedEventA_v1.kml"
    )
    json_path = vectors_path().joinpath(
        "EMSR680_AOI03_DEL_PRODUCT_observedEventA_v1.json"
    )

    # Read vectors
    kml = vectors.read(kml_path).explode(ignore_index=True)
    json = vectors.read(json_path).explode(ignore_index=True)

    # Check attributes
    assert not kml.empty
    _assert_attributes(kml, kml_path)
    _assert_attributes(json, json_path)

    # Check if equivalent
    ci.assert_geom_almost_equal(json, kml, decimal=6)

    # Just check there is no issue when opening this file
    LOGGER.debug("Open ICEYE_X2_QUICKLOOK_SC_124020_20210827T162211.kml")
    kml_path = vectors_path().joinpath(
        "ICEYE_X2_QUICKLOOK_SC_124020_20210827T162211.kml"
    )
    kml = vectors.read(kml_path)
    _assert_attributes(kml, kml_path)
    assert not kml.empty


@s3_env
def test_kmz():
    """Test KMZ files"""
    kmz_path = vectors_path().joinpath("AOI_kmz.kmz")
    gj_path = vectors_path().joinpath("AOI_kmz.geojson")

    # Read vectors
    kmz = vectors.read(kmz_path)
    gj = vectors.read(gj_path)

    # Check attributes
    _assert_attributes(kmz, kmz_path)
    _assert_attributes(gj, gj_path)

    # Check if equivalent
    assert all(
        gj.geometry.geom_equals_exact(
            kmz.to_crs(gj.crs).geometry, tolerance=0.5 * 10**-6
        )
    )


@s3_env
def test_gml():
    """Test GML functions"""
    empty_gml = vectors_path().joinpath("empty.GML")
    not_empty_gml = vectors_path().joinpath("not_empty.GML")
    naive_gml = vectors_path().joinpath("naive.GML")
    not_empty_true_path = vectors_path().joinpath("not_empty_true.geojson")

    # Empty
    empty_gdf = vectors.read(empty_gml, crs=EPSG_4326)
    assert empty_gdf.empty
    assert empty_gdf.crs == EPSG_4326
    _assert_attributes(empty_gdf, empty_gml)

    # Not empty
    not_empty_true = vectors.read(not_empty_true_path)
    not_empty = vectors.read(not_empty_gml, crs=not_empty_true.crs)
    ci.assert_geom_equal(not_empty, not_empty_true)

    # Check attrs
    _assert_attributes(not_empty_true, not_empty_true_path)
    _assert_attributes(not_empty, not_empty_gml)

    # Naive
    naive = vectors.read(naive_gml)
    assert naive.crs is None
    _assert_attributes(naive, naive_gml)


@s3_env
def test_write():
    vect_paths = [
        vectors_path().joinpath("aoi.shp"),
        vectors_path().joinpath("aoi.kml"),
        vectors_path().joinpath("aoi.geojson"),
    ]

    with tempfile.TemporaryDirectory() as tmp_dir:
        for vect_path in vect_paths:
            vect = vectors.read(vect_path)
            vect_out_path = os.path.join(tmp_dir, os.path.basename(vect_path))
            vectors.write(vect, vect_out_path)
            vect_out = vectors.read(vect_out_path)

            ci.assert_geom_equal(vect_out, vect)


@s3_env
def test_copy():
    shpfile = vectors_path().joinpath("aoi.shp")

    with tempfile.TemporaryDirectory() as tmp_dir:
        # Assert normal copy will fail
        with pytest.raises(DataSourceError):
            gpd.read_file(files.copy(shpfile, tmp_dir))

        # Assert vector copy will open in geopandas
        gpd.read_file(vectors.copy(shpfile, tmp_dir))


@s3_env
def test_utm_context_manager():
    """Test UTM context manager."""
    warning_msg = r"Geometry is in a geographic CRS. Results from 'centroid' are likely incorrect.*$"

    # Open a geographic vector
    vect = vectors.read(vectors_path().joinpath("aoi.kml"))

    # Check the vector is geographic
    assert vect.crs.is_geographic

    # Check that centroid function warns possible issues
    with pytest.warns(UserWarning, match=warning_msg):
        c = vect.centroid
        assert c.crs.is_geographic

    # Convert it
    with vectors.utm_crs(vect) as utm_vect:
        # Check the vector is projected
        assert utm_vect.crs.is_projected

        # Check no warning is now launched
        # https://docs.pytest.org/en/7.0.x/how-to/capture-warnings.html#additional-use-cases-of-warnings-in-tests
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            c2 = utm_vect.centroid
            assert c2.crs.is_projected

        # Add the centroid as a new column
        utm_vect["centroid_utm"] = c2

    # Check the CRS is back to normal
    assert vect.crs.is_geographic

    # Assert the column still exists
    assert vect["centroid_utm"].equals(c2)


def test_read_archived():
    """Test archived vectors"""
    landsat = "LM05_L1TP_200030_20121230_20200820_02_T2_CI"
    map_overlay = "map-overlay.kml"
    map_overlay_regex = ".*{}".format(map_overlay.replace(".", r"\."))
    map_overlay_extracted_path = files_path() / landsat / map_overlay
    zip_landsat = files_path() / f"{landsat}.zip"
    tar_landsat = files_path() / f"{landsat}.tar"

    map_overlay_extracted = vectors.read(map_overlay_extracted_path)

    ci.assert_geom_equal(
        map_overlay_extracted, vectors.read(f"{zip_landsat}!{landsat}/{map_overlay}")
    )
    ci.assert_geom_equal(
        map_overlay_extracted,
        vectors.read(zip_landsat, archive_regex=map_overlay_regex),
    )
    ci.assert_geom_equal(
        map_overlay_extracted,
        vectors.read(tar_landsat, archive_regex=map_overlay_regex),
    )

    file_list = path.get_archived_file_list(tar_landsat)
    ci.assert_geom_equal(
        map_overlay_extracted,
        vectors.read(tar_landsat, archive_regex=map_overlay_regex, file_list=file_list),
    )


def test_window():
    """Test read with window"""
    aoi_path = vectors_path() / "areaOfInterestA.geojson"
    vect_path = vectors_path() / "hydrographyA.geojson"

    aoi = vectors.read(aoi_path)
    vect = vectors.read(vect_path)
    vect_aoi = vectors.read(vect_path, window=aoi)
    vect_aoi_path = vectors.read(vect_path, window=aoi_path)
    vect_aoi_bounds = vectors.read(vect_path, window=aoi.total_bounds)
    vect_aoi_bbox = vectors.read(vect_path, bbox=tuple(aoi.total_bounds))

    # Tests
    ci.assert_geom_equal(vect_aoi, vect_aoi_path)
    ci.assert_geom_equal(vect_aoi, vect_aoi_bounds)
    ci.assert_geom_equal(vect_aoi, vect_aoi_bbox)

    with pytest.raises(AssertionError):
        ci.assert_geom_equal(vect, vect_aoi)

    # Manage exception by default and with reversed engine
    if vectors.is_geopandas_1_0():
        # Default engine is pyogrio, so test here fiona
        ex = TypeError
        with pytest.raises(ValueError):
            vectors.read(vect_path, bbox=aoi.total_bounds, engine="fiona")
    else:
        # Default engine is fiona, so test here pyogrio
        ex = ValueError
        with pytest.raises(TypeError):
            vectors.read(vect_path, bbox=aoi.total_bounds, engine="pyogrio")

    # Test default engine
    with pytest.raises(ex):
        vectors.read(vect_path, bbox=aoi.total_bounds)


def test_read_gdb():
    """Test read from GDB"""
    gdb_path = vectors_path() / "EMSR712_04WESERRIVERSOUTH_DelineationMap_MONIT_06.gdb"
    layer = "B1_observed_event_a"

    sertit_gdb = vectors.read(gdb_path, layer=layer)
    gpd_gdb = gpd.read_file(str(gdb_path), layer=layer)

    ci.assert_geom_equal(sertit_gdb, gpd_gdb)
    _assert_attributes(sertit_gdb, gdb_path)


def test_read_dbf():
    """Test read from GDB"""
    # DataFrame DBF (just check it works)
    dbf_path = vectors_path() / "a0_source.dbf"
    vectors.read(dbf_path)

    # GeoDataFrame DBF
    dbf_path = vectors_path() / "aoi.dbf"

    fiona = vectors.read(dbf_path, engine="fiona")
    pyogrio = vectors.read(dbf_path, engine="pyogrio")

    ci.assert_geom_equal(fiona, pyogrio)
    _assert_attributes(fiona, dbf_path)
    _assert_attributes(pyogrio, dbf_path)



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Performance module"""

import os

MAX_CORES = os.cpu_count() - 2
"""
Default maximum number of cores to use (total number of CPU minus 2).
"""

SU_MAX_CORE = "SERTIT_UTILS_MAX_CORES"
"""
Maximum core to use, default is total number of CPU minus 2.
You can update it with the environment variable :code:`SERTIT_UTILS_MAX_CORES`.
"""


def get_max_cores() -> int:
    """
    Retrieves the maximum number of CPU cores to be used by any function.

    This function determines the number of maximum allowable cores by first attempting
    to fetch the value from the environment variable `SERTIT_UTILS_MAX_CORES`. If the environment
    variable is not set, it defaults to the total number of available CPU cores minus 2.

    Returns:
        int: The maximum number of processor cores to be used.
    """
    return int(os.getenv(SU_MAX_CORE, MAX_CORES))



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import sys
from enum import unique
from functools import wraps

import tempenv

from sertit import AnyPath, ci, dask, unistra
from sertit.logs import SU_NAME
from sertit.misc import ListEnum

LOGGER = logging.getLogger(SU_NAME)

CI_SERTIT_S3 = "CI_SERTIT_USE_S3"
""" Use files stored in S3 in CI. Else uses on disk files. """

CI_SERTIT_USE_DASK = "CI_SERTIT_USE_DASK"
""" Use Dask in CI (rasters only, chunks set to auto). If not, set chunks to None. """

CI_SERTIT_TEST_LAZY = "CI_SERTIT_TEST_LAZY"
"""
Test laziness if set to 1.  
This exists because of the difficulty of creating lazy raster functions. 

WARNING: 

For now we assume function is lazy == output is still chunked.  
However, this IS abusive as some functions may compute internally and rechunk the data in the output:  
=> If a process is lazy, then its output is chunked, but the other way around is false.

How to check this behavior is not clear yet.
"""

KAPUT_KWARGS = {"fdezf": 0}

TEST_LAZY = True


@unique
class Polarization(ListEnum):
    """SAR Polarizations"""

    hh = "HH"
    vv = "VV"
    vh = "VH"
    hv = "HV"


def get_s3_ci_path():
    """Get S3 CI path"""
    unistra.define_s3_client()
    return AnyPath("s3://sertit-sertit-utils-ci")


def get_ci_data_path():
    """Get CI DATA path"""
    if int(os.getenv(CI_SERTIT_S3, 1)) and sys.platform != "win32":
        return get_s3_ci_path() / "DATA"
    else:
        return AnyPath(unistra.get_db3_path()) / "CI" / "sertit_utils" / "DATA"


def set_dask_env_var():
    if os.getenv(CI_SERTIT_USE_DASK, "1") == "1":
        os.environ[dask.SERTIT_DEFAULT_CHUNKS] = "auto"
    else:
        os.environ[dask.SERTIT_DEFAULT_CHUNKS] = "none"


def dask_env(nof_computes=None, max_computes=0):
    """
    Create dask-using environment

    Returns:
        Callable: decorated function
    """

    def dask_env_decorator(function):
        # Test laziness
        os.environ[CI_SERTIT_TEST_LAZY] = "1" if TEST_LAZY else "0"

        @wraps(function)
        def dask_env_wrapper(*_args, **_kwargs):
            """S3 environment wrapper"""
            # Not dask
            LOGGER.info("Using NUMPY")
            os.environ[CI_SERTIT_USE_DASK] = "0"
            set_dask_env_var()
            function(*_args, **_kwargs)

            # Dask
            os.environ[CI_SERTIT_USE_DASK] = "1"
            set_dask_env_var()

            with (
                dask.get_or_create_dask_client(),
                dask.raise_if_dask_computes(
                    nof_computes=nof_computes,
                    max_computes=max_computes,
                    dont_raise=False,  # os.environ[CI_SERTIT_TEST_LAZY] == "0", -> putting the right 'compute' number in nof_computes does the trick
                ),
            ):
                LOGGER.info("Using DASK multithreaded.")
                function(*_args, **_kwargs)

            with (
                dask.get_or_create_dask_client(processes=True),
                dask.raise_if_dask_computes(
                    nof_computes=nof_computes,
                    max_computes=max_computes,
                    dont_raise=False,  # os.environ[CI_SERTIT_TEST_LAZY] == "0", -> putting the right 'compute' number in nof_computes does the trick
                ),
            ):
                LOGGER.info("Using DASK with local cluster")
                function(*_args, **_kwargs)

        os.environ.pop(CI_SERTIT_USE_DASK, None)
        return dask_env_wrapper

    return dask_env_decorator


def is_not_lazy_yet(function):
    """
    The tested fiunction is not lazy yet : don't evaluate it

    Returns:
        Callable: decorated function
    """

    @wraps(function)
    def is_not_lazy_yet_wrapper(*_args, **_kwargs):
        """S3 environment wrapper"""
        with tempenv.TemporaryEnvironment({CI_SERTIT_TEST_LAZY: "0"}):
            LOGGER.warning(
                "This function is not lazy yet. Laziness won't be tested here."
            )
            function(*_args, **_kwargs)

    return is_not_lazy_yet_wrapper


def rasters_path():
    return get_ci_data_path().joinpath("rasters")


def vectors_path():
    return get_ci_data_path().joinpath("vectors")


def geometry_path():
    return get_ci_data_path().joinpath("geometry")


def files_path():
    return get_ci_data_path().joinpath("files")


def display_path():
    return get_ci_data_path().joinpath("display")


def xml_path():
    return get_ci_data_path().joinpath("xml")


def s3_env(*args, **kwargs):
    return unistra.s3_env(use_s3_env_var=CI_SERTIT_S3, *args, **kwargs)  # noqa: B026


def get_output(tmp, file, debug=False, dask_folders=False):
    if debug:
        out_path = AnyPath(__file__).resolve().parent / "ci_output"

        if dask_folders:
            client = dask.get_client()
            if client:
                multithreading = not client.cluster.processes
                if multithreading:
                    out_path /= "dask_multithreading"
                else:
                    out_path /= "dask_multiple_clusters"
            else:
                out_path /= "numpy"

        out_path.mkdir(parents=True, exist_ok=True)
        return out_path / file
    else:
        return AnyPath(tmp, file)


def assert_chunked_computed(result, text):
    """ """

    if os.environ[CI_SERTIT_TEST_LAZY] == "1":
        if os.environ[CI_SERTIT_USE_DASK] == "1":
            ci.assert_chunked(result), f"{text}: Your data should be chunked!"
        else:
            ci.assert_computed(result), f"{text}: Your data should be computed!"
    else:
        if os.environ[CI_SERTIT_USE_DASK] == "1" and not dask.is_chunked(result):
            LOGGER.error(
                f"{text}: You are currently using dask and therefore your function should be chunked. However, your output has been computed."
            )
        elif os.environ[CI_SERTIT_USE_DASK] == "0" and dask.is_chunked(result):
            LOGGER.error(
                f"{text}: You are currently using numpy and therefore your function should load data into memory. However, your output is lazy."
            )



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing the files"""

import os
import shutil
import tempfile
from datetime import date, datetime

import numpy as np
import pytest
from lxml import etree, html

from ci.script_utils import Polarization, files_path, s3_env
from sertit import AnyPath, ci, files, path, vectors

ci.reduce_verbosity()


def test_archive():
    """Test extracting functions"""
    with tempfile.TemporaryDirectory() as tmp_dir:
        # Archives
        zip_file = files_path().joinpath("test_zip.zip")
        zip2_file = files_path().joinpath("test_zip.zip")  # For overwrite
        zip_without_directory = files_path().joinpath("test_zip_without_directory.zip")
        tar_file = files_path().joinpath("test_tar.tar")
        tar_gz_file = files_path().joinpath("test_targz.tar.gz")

        # Core dir
        core_dir = files_path().joinpath("core")
        folder = core_dir
        archives = [
            zip_file,
            tar_file,
            tar_gz_file,
            folder,
            zip2_file,
            zip_without_directory,
        ]

        # Extract
        extracted_dirs = files.extract_files(archives, tmp_dir, overwrite=True)
        files.extract_files([zip2_file], tmp_dir, overwrite=False)  # Already existing

        # Test
        for ex_dir in extracted_dirs:
            ci.assert_dir_equal(core_dir, ex_dir)

        # Archive
        archive_base = os.path.join(tmp_dir, "archive")
        for fmt in ["zip", "tar", "gztar"]:
            archive_fn = files.archive(
                folder_path=core_dir, archive_path=archive_base, fmt=fmt
            )
            out = files.extract_file(archive_fn, tmp_dir)
            # an additional folder is created
            out_dir = path.listdir_abspath(out)[0]
            ci.assert_dir_equal(core_dir, out_dir)

            # Remove out directory in order to avoid any interferences
            files.remove(out)

        # Add to zip
        zip_out = zip2_file if path.is_cloud_path(zip2_file) else archive_base + ".zip"
        core_copy = files.copy(core_dir, os.path.join(tmp_dir, "core2"))
        zip_out = files.add_to_zip(zip_out, core_copy)

        # Extract
        unzip_out = os.path.join(tmp_dir, "out")
        unzip_out = files.extract_file(zip_out, unzip_out)

        # Test
        unzip_dirs = path.listdir_abspath(unzip_out)

        assert len(unzip_dirs) == 2
        ci.assert_dir_equal(unzip_dirs[0], unzip_dirs[1])


@s3_env
def test_archived_files():
    landsat_name = "LM05_L1TP_200030_20121230_20200820_02_T2_CI"
    ok_folder = files_path().joinpath(landsat_name)
    zip_file = files_path().joinpath(f"{landsat_name}.zip")
    tar_file = files_path().joinpath(f"{landsat_name}.tar")
    targz_file = files_path().joinpath(f"{landsat_name}.tar.gz")
    sz_file = files_path().joinpath(f"{landsat_name}.7z")

    # VECTORS
    vect_name = "map-overlay.kml"
    vec_ok_path = ok_folder.joinpath(vect_name)
    if shutil.which("ogr2ogr"):  # Only works if ogr2ogr can be found.
        vect_regex = f".*{vect_name}"
        vect_zip = vectors.read(zip_file, archive_regex=vect_regex)
        vect_tar = vectors.read(tar_file, archive_regex=r".*overlay\.kml")
        vect_ok = vectors.read(vec_ok_path)
        assert not vect_ok.empty
        ci.assert_geom_equal(vect_ok, vect_zip)
        ci.assert_geom_equal(vect_ok, vect_tar)

    with tempfile.TemporaryDirectory() as tmp_dir:
        # XML
        xml_name = "LM05_L1TP_200030_20121230_20200820_02_T2_MTL.xml"
        xml_ok_path = ok_folder.joinpath(xml_name)
        if path.is_cloud_path(files_path()):
            xml_ok_path = str(xml_ok_path.download_to(tmp_dir))
        else:
            xml_ok_path = str(xml_ok_path)

        xml_regex = f".*{xml_name}"
        xml_zip = files.read_archived_xml(zip_file, xml_regex)
        xml_tar = files.read_archived_xml(tar_file, r".*_MTL\.xml")
        xml_ok = etree.parse(xml_ok_path).getroot()
        ci.assert_xml_equal(xml_ok, xml_zip)
        ci.assert_xml_equal(xml_ok, xml_tar)

        # FILE + HTML
        html_zip_file = files_path().joinpath("productPreview.zip")
        html_tar_file = files_path().joinpath("productPreview.tar")
        html_name = "productPreview.html"
        html_ok_path = files_path().joinpath(html_name)
        if path.is_cloud_path(files_path()):
            html_ok_path = str(html_ok_path.download_to(tmp_dir))
        else:
            html_ok_path = str(html_ok_path)

        html_regex = f".*{html_name}"

        # FILE
        file_zip = files.read_archived_file(html_zip_file, html_regex)
        file_tar = files.read_archived_file(html_tar_file, html_regex)
        html_ok = html.parse(html_ok_path).getroot()
        ci.assert_html_equal(html_ok, html.fromstring(file_zip))
        ci.assert_html_equal(html_ok, html.fromstring(file_tar))

        file_list = path.get_archived_file_list(html_zip_file)
        ci.assert_html_equal(
            html_ok,
            html.fromstring(
                files.read_archived_file(html_zip_file, html_regex, file_list=file_list)
            ),
        )

        # HTML
        html_zip = files.read_archived_html(html_zip_file, html_regex)
        html_tar = files.read_archived_html(html_tar_file, html_regex)
        ci.assert_html_equal(html_ok, html_zip)
        ci.assert_html_equal(html_ok, html_tar)
        ci.assert_html_equal(
            html_ok,
            files.read_archived_html(
                html_tar_file,
                html_regex,
                file_list=path.get_archived_file_list(html_tar_file),
            ),
        )

        # ERRORS
        with pytest.raises(TypeError):
            files.read_archived_file(targz_file, xml_regex)
        with pytest.raises(TypeError):
            files.read_archived_file(sz_file, xml_regex)
        with pytest.raises(FileNotFoundError):
            files.read_archived_file(zip_file, "cdzeferf")


def test_cp_rm():
    """Test CP/RM functions"""
    with tempfile.TemporaryDirectory() as tmp_dir:
        empty_tmp = os.listdir(tmp_dir)

        # Copy file
        curr_path = os.path.realpath(__file__)
        file_1 = files.copy(curr_path, tmp_dir)
        file_2 = files.copy(curr_path, os.path.join(tmp_dir, "test_pattern.py"))

        # Copy dir
        dir_path = os.path.dirname(curr_path)
        test_dir = files.copy(
            dir_path, os.path.join(tmp_dir, os.path.basename(dir_path))
        )

        # Test copy
        assert os.path.isfile(file_1)
        assert os.path.isfile(file_2)
        assert os.path.isdir(test_dir)

        # Remove file
        files.remove(file_1)
        files.remove("non_existing_file.txt")
        files.remove_by_pattern(tmp_dir, name_with_wildcard="*pattern*", extension="py")

        # Remove dir
        files.remove(test_dir)

        # Assert tempfile is empty
        assert os.listdir(tmp_dir) == empty_tmp


def test_json():
    """Test json functions"""

    test_dict = {
        "A": 3,
        "C": "m2",  # Can be parsed as a date, we do not want that !
        "D": datetime.today(),
        "Dbis": date.today(),
        "E": np.int64(15),
        "F": Polarization.vv,
        "G": True,
        "H": AnyPath("/home/data"),
    }

    with tempfile.TemporaryDirectory() as tmp_dir:
        json_file = os.path.join(tmp_dir, "test.json")

        # Save JSON
        files.save_json(test_dict, json_file)

        # Load JSON
        obj = files.read_json(json_file)

        assert (
            obj.pop("F") == test_dict.pop("F").value
        )  # Enum are stored following their value

        assert obj.pop("H") == str(
            test_dict.pop("H")
        )  # Enum are stored following their value
        assert obj == test_dict

        # Test deprecation
        with pytest.deprecated_call():
            files.save_json(json_file, test_dict)


def test_pickle():
    """Test pickle functions"""

    test_dict = {
        "A": 3,
        "B": np.zeros((3, 3)),
        "C": "str",
        "D": datetime.today(),
        "E": np.int64(15),
    }

    with tempfile.TemporaryDirectory() as tmp_dir:
        pkl_file = os.path.join(tmp_dir, "test.pkl")

        # Save pickle
        files.save_obj(test_dict, pkl_file)

        # Load pickle
        obj = files.load_obj(pkl_file)

        # Test (couldn't compare the dicts as they contain numpy arrays)
        np.testing.assert_equal(obj, test_dict)


def test_hash_file_content():
    """Test hash_file_content"""
    file_content = "This is a test."

    # Test
    hashed = files.hash_file_content(file_content)

    # Test
    assert hashed == "16c5bf1fc5"
